# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import json
from pathlib import Path
from typing import IO, Self

from pydiverse.colspec.pdt_util import num_rows

from .config import Config, alias_subquery
from .optional_dependency import ColExpr, pdt, pl


class FailureInfo:
    """A container carrying information about rows failing validation in
    :meth:`ColSpec.filter`."""

    def __init__(
        self,
        tbl: pdt.Table,
        invalid_rows: pdt.Table,
        rule_columns: dict[str, ColExpr],
        cfg: Config,
    ):
        #: The subset of the input data frame containing the *invalid* rows along with
        # all boolean columns used for validation. Each of these boolean columns
        # describes a single rule where a value of ``False``` indicates unsuccessful
        # validation. Thus, at least one value per row is ``False``.
        self.tbl = tbl
        self._invalid_rows = invalid_rows >> alias_subquery(cfg=cfg)
        #: The columns in `_tbl` which are used for validation.
        self.rule_columns = rule_columns
        self.cfg = cfg

    @property
    def invalid_rows(self) -> pdt.Table:
        from pydiverse.transform.extended import select

        return self._invalid_rows >> select(*[c.name for c in self.tbl])

    @property
    def debug_invalid_rows(self) -> pdt.Table:
        return self._invalid_rows

    def counts(self) -> dict[str, int]:
        """The number of validation failures for each individual rule.

        Returns:
            A mapping from rule name to counts. If a rule's failure count is 0, it is
            not included here.
        """
        if not self.rule_columns:  # pdt summarize needs at least one column
            return {}
        from pydiverse.transform.extended import C, export, summarize

        # There can be nulls, which should not be counted.
        cnts: dict[str, int] = (
            self._invalid_rows
            >> summarize(**{k: (~C[k]).sum() for k in self.rule_columns.keys()})
            >> export(pdt.Dict())
        )

        return {k: v for k, v in cnts.items() if v is not None and v > 0}

    def __len__(self) -> int:
        return num_rows(self._invalid_rows)

    # ---------------------------------- PERSISTENCE --------------------------------- #

    def write_parquet(self, file: str | Path | IO[bytes]):
        """Write the failure info to a Parquet file.

        Args:
            file: The file path or writable file-like object to write to.
        """
        # NOTE: We add a dummy column with metadata in the column name to allow writing
        #  the rule columns to the same file.
        rule_columns_json = json.dumps({"rule_columns": self.rule_columns})
        (self._invalid_rows >> pdt.export(pdt.Polars())).with_columns(
            pl.lit(None).alias(rule_columns_json),
        ).write_parquet(file)

    @classmethod
    def scan_parquet(cls, source: str | Path | IO[bytes]) -> Self:
        """Lazily read the parquet file with the failure info.

        Args:
            source: The file path or readable file-like object to read from.

        Returns:
            The failure info object.
        """
        lf = pl.scan_parquet(source)
        # NOTE: In `write_parquet`, the rule columns are added as the name of the last
        #  column.
        last_column = lf.collect_schema().names()[-1]
        rule_columns = json.loads(last_column)["rule_columns"]
        return cls(
            pdt.Table(pl.DataFrame()),
            pdt.Table(lf.drop(last_column)),
            rule_columns,
            Config.default,
        )


# ------------------------------------ COMPUTATION ----------------------------------- #


def _compute_counts(df: pl.DataFrame, rule_columns: list[str]) -> dict[str, int]:
    if len(rule_columns) == 0:
        return {}

    counts = df.select((~pl.col(rule_columns)).sum())
    return {
        name: count for name, count in (counts.row(0, named=True).items()) if count > 0
    }


def _compute_cooccurrence_counts(
    df: pl.DataFrame, rule_columns: list[str]
) -> dict[frozenset[str], int]:
    if len(rule_columns) == 0:
        return {}

    group_lengths = df.group_by(pl.col(rule_columns).fill_null(True)).len()
    if len(group_lengths) == 0:
        return {}

    groups = group_lengths.drop("len")
    counts = group_lengths.get_column("len")
    return {
        frozenset(
            name
            for name, success in zip(rule_columns, row, strict=False)
            if not success
        ): count
        for row, count in zip(groups.iter_rows(), counts, strict=False)
    }



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

import pydiverse.colspec as cs
import pydiverse.colspec.collection
from pydiverse.colspec.optional_dependency import dy, pl


class MyColSpec(cs.ColSpec):
    a = cs.Int64()


class SimpleCollection(pydiverse.colspec.collection.Collection):
    first: MyColSpec
    second: MyColSpec | None
    third: MyColSpec | None


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_concat():
    col1 = SimpleCollection.cast_polars_data({"first": pl.LazyFrame({"a": [1, 2, 3]})})
    col2 = SimpleCollection.cast_polars_data(
        {
            "first": pl.LazyFrame({"a": [4, 5, 6]}),
            "second": pl.LazyFrame({"a": [4, 5, 6]}),
        }
    )
    col3 = SimpleCollection.cast_polars_data(
        {
            "first": pl.LazyFrame({"a": [7, 8, 9]}),
            "second": pl.LazyFrame({"a": [7, 8, 9]}),
            "third": pl.LazyFrame({"a": [7, 8, 9]}),
        }
    )
    concat = dy.concat_collection_members([col1, col2, col3])
    assert concat["first"].collect().get_column("a").to_list() == list(range(1, 10))
    assert concat["second"].collect().get_column("a").to_list() == list(range(4, 10))
    assert concat["third"].collect().get_column("a").to_list() == list(range(7, 10))


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_concat_empty():
    with pytest.raises(ValueError):
        dy.concat_collection_members([])



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

import pydiverse.colspec as cs
from pydiverse.colspec.columns._utils import pydiverse_type_opinions
from pydiverse.colspec.optional_dependency import dy, pa
from pydiverse.colspec.testing import (
    ALL_COLUMN_TYPES,
    COLUMN_TYPES,
    SUPERTYPE_COLUMN_TYPES,
    create_colspec,
)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize("column_type", set(ALL_COLUMN_TYPES) - {cs.Any})
def test_equal_to_polars_schema(column_type: type[cs.Column]):
    schema = create_colspec("test", {"a": column_type()})
    actual = schema.pyarrow_schema()
    expected = schema.create_empty_polars().to_arrow().schema

    actual_dict = {f.name: f.type for f in actual}

    expected_dict = {f.name: pydiverse_type_opinions(f.type) for f in expected}
    assert actual_dict == expected_dict


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_equal_polars_schema_enum():
    schema = create_colspec("test", {"a": cs.Enum(["a", "b"])})
    actual = schema.pyarrow_schema()
    expected = schema.create_empty_polars().to_arrow().schema
    assert actual == expected


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize(
    "inner",
    [c() for c in set(ALL_COLUMN_TYPES) - {cs.Any, cs.Enum}]
    + [cs.List(t()) for t in set(ALL_COLUMN_TYPES) - {cs.Any, cs.Enum}]
    + [cs.Struct({"a": t()}) for t in set(ALL_COLUMN_TYPES) - {cs.Any, cs.Enum}],
)
def test_equal_polars_schema_list(inner: cs.Column):
    schema = create_colspec("test", {"a": cs.List(inner)})
    actual = schema.pyarrow_schema()
    expected = schema.create_empty_polars().to_arrow().schema
    # replace large_list with list
    expected = pa.schema(
        [
            pa.field(
                expected[0].name,
                pydiverse_type_opinions(expected[0].type),
                nullable=expected[0].nullable,
            )
        ]
    )
    assert actual == expected


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize(
    "inner",
    [c() for c in ALL_COLUMN_TYPES]
    + [cs.Struct({"a": t()}) for t in ALL_COLUMN_TYPES]
    + [cs.List(t()) for t in ALL_COLUMN_TYPES],
)
def test_equal_polars_schema_struct(inner: cs.Column):
    schema = create_colspec("test", {"a": cs.Struct({"a": inner})})
    actual = schema.pyarrow_schema()
    expected = schema.create_empty_polars().to_arrow().schema
    assert actual == expected


@pytest.mark.skipif(pa.Field is None, reason="pyarrow is required for this test")
@pytest.mark.parametrize("column_type", COLUMN_TYPES + SUPERTYPE_COLUMN_TYPES)
@pytest.mark.parametrize("nullable", [True, False])
def test_nullability_information(column_type: type[cs.Column], nullable: bool):
    schema = create_colspec("test", {"a": column_type(nullable=nullable)})
    assert ("not null" in str(schema.pyarrow_schema())) != nullable


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize("nullable", [True, False])
def test_nullability_information_enum(nullable: bool):
    schema = create_colspec("test", {"a": cs.Enum(["a", "b"], nullable=nullable)})
    assert ("not null" in str(schema.pyarrow_schema())) != nullable


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize(
    "inner",
    [c() for c in ALL_COLUMN_TYPES]
    + [cs.List(t()) for t in ALL_COLUMN_TYPES]
    + [cs.Struct({"a": t()}) for t in ALL_COLUMN_TYPES],
)
@pytest.mark.parametrize("nullable", [True, False])
def test_nullability_information_list(inner: cs.Column, nullable: bool):
    schema = create_colspec("test", {"a": cs.List(inner, nullable=nullable)})
    assert ("not null" in str(schema.pyarrow_schema())) != nullable


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize(
    "inner",
    [c() for c in ALL_COLUMN_TYPES]
    + [cs.Struct({"a": t()}) for t in ALL_COLUMN_TYPES]
    + [cs.List(t()) for t in ALL_COLUMN_TYPES],
)
@pytest.mark.parametrize("nullable", [True, False])
def test_nullability_information_struct(inner: cs.Column, nullable: bool):
    schema = create_colspec("test", {"a": cs.Struct({"a": inner}, nullable=nullable)})
    assert ("not null" in str(schema.pyarrow_schema())) != nullable


@pytest.mark.skipif(pa.Field is None, reason="pyarrow is required for this test")
def test_multiple_columns():
    schema = create_colspec("test", {"a": cs.Int32(nullable=False), "b": cs.Integer()})
    assert str(schema.pyarrow_schema()).split("\n") == ["a: int32 not null", "b: int64"]



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

import sys
from typing import Any

import pytest

import pydiverse.colspec as cs
from pydiverse.colspec.columns.float import _BaseFloat
from pydiverse.colspec.optional_dependency import (
    FLOAT_DTYPES,
    INTEGER_DTYPES,
    C,
    DataTypeClass,
    dy,
    pdt,
    pl,
)
from pydiverse.colspec.testing import FLOAT_COLUMN_TYPES
from pydiverse.colspec.testing.rules import evaluate_rules


class FloatColSpec(cs.ColSpec):
    a = cs.Float()


@pytest.mark.parametrize("column_type", FLOAT_COLUMN_TYPES)
@pytest.mark.parametrize(
    "kwargs",
    [
        {"min": 2, "max": 1},
        {"min_exclusive": 2, "max": 2},
        {"min": 2, "max_exclusive": 2},
        {"min_exclusive": 2, "max_exclusive": 2},
        {"min": 2, "min_exclusive": 2},
        {"max": 2, "max_exclusive": 2},
    ],
)
def test_args_consistency_min_max(
    column_type: type[_BaseFloat], kwargs: dict[str, Any]
):
    with pytest.raises(ValueError):
        column_type(**kwargs)


@pytest.mark.parametrize(
    ("column_type", "kwargs"),
    [
        (cs.Float, dict(min=float("-inf"))),
        (cs.Float, dict(max=float("inf"))),
        (cs.Float32, dict(min=-sys.float_info.max)),
        (cs.Float32, dict(max=sys.float_info.max)),
        (cs.Float64, dict(min=float("-inf"))),
        (cs.Float64, dict(max=float("inf"))),
    ],
)
def test_invalid_args(column_type: type[_BaseFloat], kwargs: dict[str, Any]):
    with pytest.raises(ValueError):
        column_type(**kwargs)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize("dtype", FLOAT_DTYPES)
def test_any_integer_dtype_passes(dtype: DataTypeClass):
    df = pl.DataFrame(schema={"a": dtype})
    assert FloatColSpec.is_valid_polars(df)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize("dtype", [pl.Boolean, pl.String] + list(INTEGER_DTYPES))
def test_non_integer_dtype_fails(dtype: DataTypeClass):
    df = pl.DataFrame(schema={"a": dtype})
    assert not FloatColSpec.is_valid_polars(df)


@pytest.mark.skipif(C is None, reason="pydiverse-transform is required for this test")
@pytest.mark.parametrize("column_type", FLOAT_COLUMN_TYPES)
@pytest.mark.parametrize(
    ("inclusive", "valid"),
    [
        (True, {"min": [False, False, True, True, True]}),
        (False, {"min_exclusive": [False, False, False, True, True]}),
    ],
)
def test_validate_min(
    column_type: type[_BaseFloat], inclusive: bool, valid: dict[str, list[bool]]
):
    kwargs = {("min" if inclusive else "min_exclusive"): 3}
    column = column_type(**kwargs)  # type: ignore
    tbl = pdt.Table({"a": [1, 2, 3, 4, 5]})
    actual = evaluate_rules(tbl, column.validation_rules(tbl.a))
    assert actual == valid


@pytest.mark.skipif(C is None, reason="pydiverse-transform is required for this test")
@pytest.mark.parametrize("column_type", FLOAT_COLUMN_TYPES)
@pytest.mark.parametrize(
    ("inclusive", "valid"),
    [
        (True, {"max": [True, True, True, False, False]}),
        (False, {"max_exclusive": [True, True, False, False, False]}),
    ],
)
def test_validate_max(
    column_type: type[_BaseFloat], inclusive: bool, valid: dict[str, list[bool]]
):
    kwargs = {("max" if inclusive else "max_exclusive"): 3}
    column = column_type(**kwargs)  # type: ignore
    tbl = pdt.Table({"a": [1, 2, 3, 4, 5]})
    actual = evaluate_rules(tbl, column.validation_rules(tbl.a))
    assert actual == valid


@pytest.mark.skipif(C is None, reason="pydiverse-transform is required for this test")
@pytest.mark.parametrize("column_type", FLOAT_COLUMN_TYPES)
@pytest.mark.parametrize(
    ("min_inclusive", "max_inclusive", "valid"),
    [
        (
            True,
            True,
            {
                "min": [False, True, True, True, True],
                "max": [True, True, True, True, False],
            },
        ),
        (
            True,
            False,
            {
                "min": [False, True, True, True, True],
                "max_exclusive": [True, True, True, False, False],
            },
        ),
        (
            False,
            True,
            {
                "min_exclusive": [False, False, True, True, True],
                "max": [True, True, True, True, False],
            },
        ),
        (
            False,
            False,
            {
                "min_exclusive": [False, False, True, True, True],
                "max_exclusive": [True, True, True, False, False],
            },
        ),
    ],
)
def test_validate_range(
    column_type: type[_BaseFloat],
    min_inclusive: bool,
    max_inclusive: bool,
    valid: dict[str, list[bool]],
):
    kwargs = {
        ("min" if min_inclusive else "min_exclusive"): 2,
        ("max" if max_inclusive else "max_exclusive"): 4,
    }
    column = column_type(**kwargs)  # type: ignore
    tbl = pdt.Table({"a": [1, 2, 3, 4, 5]})
    actual = evaluate_rules(tbl, column.validation_rules(tbl.a))
    assert actual == valid


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_sample_unchecked_min_0():
    column = dy.Float(min=0, max=10)
    actual = column._sample_unchecked(dy.random.Generator(), n=10000)
    assert actual.min() >= 0, "There should be no negative values"  # type: ignore



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

import pydiverse.colspec as cs


class ParentColSpec(cs.ColSpec):
    a = cs.Integer()


class ChildColSpec(ParentColSpec):
    b = cs.Integer()


class GrandchildColSpec(ChildColSpec):
    c = cs.Integer()


def test_columns():
    assert ParentColSpec.column_names() == ["a"]
    assert ChildColSpec.column_names() == ["a", "b"]
    assert GrandchildColSpec.column_names() == ["a", "b", "c"]



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

# NOTE: This file does not actually run any tests. Instead, it calls functions for which
# we simply want to ensure that our type checking works as desired. In some instances,
# we add 'type: ignore' markers here but, paired with "warn_unused_ignores = true", this
# allows testing that typing fails where we want it to without failing pre-commit
# checks.

import datetime
import decimal
import functools
from dataclasses import dataclass
from typing import Any

import pytest

import pydiverse.colspec as cs
import pydiverse.colspec.collection
from pydiverse.colspec import ColSpec
from pydiverse.colspec.columns import ColExpr
from pydiverse.colspec.optional_dependency import C, dy, pl

# Note: To properly test the typing of the library,
# we also need to make sure that imported colspecs are properly processed.
from pydiverse.colspec.testing.typing import MyImportedColSpec

# pytestmark = pytest.mark.skip(reason="typing-only tests")

#                                        FRAMES                                        #
# ------------------------------------------------------------------------------------ #


class ColumnSpecification(ColSpec):
    a = cs.Int64()


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_data_frame_lazy():
    df = ColumnSpecification.create_empty_polars()
    df2 = df.lazy()
    ColumnSpecification.validate_polars(df2)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_lazy_frame_lazy():
    df = ColumnSpecification.create_empty_polars().lazy()
    df2 = df.lazy()
    ColumnSpecification.validate_polars(df2)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_lazy_frame_collect():
    df = ColumnSpecification.create_empty_polars().lazy()
    df2 = df.collect()
    ColumnSpecification.validate_polars(df2)


# ------------------------------------------------------------------------------------ #
#                                      COLLECTION                                      #
# ------------------------------------------------------------------------------------ #


class MyFirstColSpec(ColSpec):
    a = cs.Integer(primary_key=True)


class MySecondColSpec(ColSpec):
    a = cs.Integer(primary_key=True)
    b = cs.Integer()


@dataclass
class MyCollection(pydiverse.colspec.collection.Collection):
    first: MyFirstColSpec
    second: MySecondColSpec


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_collection_filter_return_value():
    _, failure = MyCollection.filter_polars_data(
        {
            "first": MyFirstColSpec.sample_polars(3),
            "second": MySecondColSpec.sample_polars(2),
        },
    )
    assert len(failure["first"]) == 0  # type: ignore[misc]


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_collection_filter_return_value2():
    c = MyCollection(
        first=MyFirstColSpec.sample_polars(3), second=MySecondColSpec.sample_polars(2)
    )
    _, failure = c.filter_polars()
    assert len(failure["first"]) == 0  # type: ignore[misc]


# ------------------------------------------------------------------------------------ #
#                                       ITER ROWS                                      #
# ------------------------------------------------------------------------------------ #


Char = functools.partial(cs.String, min_length=1, max_length=1)
Flags = functools.partial(cs.Struct, inner={"x": Char(), "y": Char()})


class MyColSpec(ColSpec):
    a = cs.Int64()
    b = cs.Float32()
    c = cs.Enum(["a", "b", "c"])
    d = cs.Struct({"a": cs.Int64(), "b": cs.Struct({"c": cs.Enum(["a", "b"])})})
    e = cs.List(cs.Struct({"a": cs.Int64()}))
    f = cs.Datetime()
    g = cs.Date()
    h = cs.Any()
    some_decimal = cs.Decimal(12, 8)
    custom_col = Flags()
    custom_col_list = cs.List(Flags())

    @cs.rule()
    @staticmethod
    def b_greater_a() -> ColExpr:
        return C.b > C.a


@pytest.fixture
def my_colspec_df() -> MyColSpec:
    return MyColSpec.validate_polars(
        pl.DataFrame(
            {
                "a": [1],
                "b": [1.0],
                "c": ["a"],
                "d": [{"a": 1, "b": {"c": "a"}}],
                "e": [[{"a": 1}]],
                "f": [datetime.datetime(2022, 1, 1, 0, 0, 0)],
                "g": [datetime.date(2022, 1, 1)],
                "h": [1],
                "some_decimal": [decimal.Decimal("1.5")],
                "custom_col": [{"x": "a", "y": "b"}],
                "custom_col_list": [[{"x": "a", "y": "b"}]],
            }
        ),
        cast=True,
    )


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_iter_rows_assignment_correct_type(my_colspec_df: pl.DataFrame):
    entry = next(my_colspec_df.iter_rows(named=True))

    a: int = entry["a"]  # noqa: F841
    b: Any = entry["custom_col"]  # noqa: F841
    c: list[Any] = entry["custom_col_list"]  # noqa: F841


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_iter_rows_colspec_subtypes(my_colspec_df: pl.DataFrame):
    class MySubColSpec(MyColSpec):
        i = cs.Int64()

    class MySubSubColSpec(MySubColSpec):
        j = cs.Int64()

    my_sub_colspec_df = MySubColSpec.validate_polars(
        my_colspec_df.with_columns(i=pl.lit(2, dtype=pl.Int64))
    )
    entry1 = next(my_sub_colspec_df.iter_rows(named=True))

    a1: int = entry1["a"]  # noqa: F841
    i1: int = entry1["i"]  # noqa: F841

    my_sub_sub_colspec_df = MySubSubColSpec.validate_polars(
        my_sub_colspec_df.with_columns(j=pl.lit(2, dtype=pl.Int64))
    )
    entry2 = next(my_sub_sub_colspec_df.iter_rows(named=True))

    a2: int = entry2["a"]  # noqa: F841
    i2: int = entry2["i"]  # noqa: F841
    j2: int = entry2["j"]  # noqa: F841


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_iter_rows_assignment_wrong_type(my_colspec_df: dy.DataFrame[MyColSpec]):
    entry = next(my_colspec_df.iter_rows(named=True))

    a: int = entry["b"]  # type: ignore[assignment] # noqa: F841


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_iter_rows_read_only(my_colspec_df: dy.DataFrame[MyColSpec]):
    entry = next(my_colspec_df.iter_rows(named=True))

    entry["a"] = 1  # type: ignore[typeddict-readonly-mutated]


# @pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
# def test_iter_rows_missing_key(my_colspec_df: dy.DataFrame[MyColSpec]):
#     entry = next(my_colspec_df.iter_rows(named=True))
#
#     _ = entry["i"]  # type: ignore[misc]


# @pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
# def test_iter_rows_without_named(my_colspec_df: dy.DataFrame[MyColSpec]):
#     # Make sure we don't accidentally override the return type of `iter_rows` with
#     # `named=False`.
#     entry = next(my_colspec_df.iter_rows(named=False))
#
#     _ = entry["g"]  # type: ignore[call-overload]


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_iter_rows_imported_colspec():
    my_imported_colspec_df = MyImportedColSpec.validate_polars(
        pl.DataFrame(
            {
                "a": [1],
                "b": [1.0],
                "c": ["a"],
                "d": [{"a": 1, "b": {"c": "a"}}],
                "e": [[{"a": 1}]],
                "f": [datetime.datetime(2022, 1, 1, 0, 0, 0)],
                "g": [datetime.date(2022, 1, 1)],
                "h": [1],
                "some_decimal": [decimal.Decimal("1.5")],
            }
        ),
        cast=True,
    )
    entry = next(my_imported_colspec_df.iter_rows(named=True))

    a: int = entry["a"]  # noqa: F841
    b: int = entry["b"]  # type: ignore[assignment] # noqa: F841
    entry["a"] = 1  # type: ignore[typeddict-readonly-mutated]
    # _ = entry["i"]  # type: ignore[misc]


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_iter_rows_imported_subcolspec():
    class MySubFromImportedColSpec(MyImportedColSpec):
        i = cs.Int64()

    my_sub_from_imported_colspec_df = MySubFromImportedColSpec.validate_polars(
        pl.DataFrame(
            {
                "a": [1],
                "b": [1.0],
                "c": ["a"],
                "d": [{"a": 1, "b": {"c": "a"}}],
                "e": [[{"a": 1}]],
                "f": [datetime.datetime(2022, 1, 1, 0, 0, 0)],
                "g": [datetime.date(2022, 1, 1)],
                "h": [1],
                "some_decimal": [decimal.Decimal("1.5")],
                "i": [1],
            }
        ),
        cast=True,
    )
    entry = next(my_sub_from_imported_colspec_df.iter_rows(named=True))

    _ = entry["i"]  # noqa: F841



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

import decimal
from typing import Any

import pytest

import pydiverse.colspec as cs
from pydiverse.colspec.optional_dependency import (
    FLOAT_DTYPES,
    INTEGER_DTYPES,
    C,
    DataTypeClass,
    pdt,
    pl,
)
from pydiverse.colspec.testing.rules import evaluate_rules


class DecimalColSpec(cs.ColSpec):
    a = cs.Decimal()


@pytest.mark.parametrize(
    "kwargs",
    [
        {"min": decimal.Decimal(2), "max": decimal.Decimal(1)},
        {"min_exclusive": decimal.Decimal(2), "max": decimal.Decimal(2)},
        {"min": decimal.Decimal(2), "max_exclusive": decimal.Decimal(2)},
        {"min_exclusive": decimal.Decimal(2), "max_exclusive": decimal.Decimal(2)},
        {"min": decimal.Decimal(2), "min_exclusive": decimal.Decimal(2)},
        {"max": decimal.Decimal(2), "max_exclusive": decimal.Decimal(2)},
    ],
)
def test_args_consistency_min_max(kwargs: dict[str, Any]):
    with pytest.raises(ValueError):
        cs.Decimal(**kwargs)


@pytest.mark.parametrize(
    "kwargs",
    [
        dict(scale=1, min=decimal.Decimal("3.14")),
        dict(scale=1, min_exclusive=decimal.Decimal("3.14")),
        dict(scale=1, max=decimal.Decimal("3.14")),
        dict(scale=1, max_exclusive=decimal.Decimal("3.14")),
        dict(min=decimal.Decimal(float("inf"))),
        dict(max=decimal.Decimal(float("inf"))),
        dict(scale=0, precision=2, min=decimal.Decimal("100")),
        dict(scale=0, precision=2, max=decimal.Decimal("100")),
    ],
)
def test_invalid_args(kwargs: dict[str, Any]):
    with pytest.raises(ValueError):
        cs.Decimal(**kwargs)


@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
@pytest.mark.parametrize(
    "dtype", [pl.Decimal, pl.Decimal(12), pl.Decimal(None, 8), pl.Decimal(6, 2)]
)
def test_any_decimal_dtype_passes(dtype: DataTypeClass):
    df = pl.DataFrame(schema={"a": dtype})
    tbl = pdt.Table(df)
    assert DecimalColSpec.is_valid(tbl)


@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
@pytest.mark.parametrize(
    "dtype", [pl.Boolean, pl.String] + list(INTEGER_DTYPES) + list(FLOAT_DTYPES)
)
def test_non_decimal_dtype_fails(dtype: DataTypeClass):
    if dtype == pl.Int128:
        # this type is not supported by pydiverse libraries, yet
        return
    df = pl.DataFrame(schema={"a": dtype})
    tbl = pdt.Table(df)
    assert not DecimalColSpec.is_valid(tbl)


@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
@pytest.mark.parametrize(
    ("inclusive", "valid"),
    [
        (True, {"min": [False, False, True, True, True]}),
        (False, {"min_exclusive": [False, False, False, True, True]}),
    ],
)
def test_validate_min(inclusive: bool, valid: dict[str, list[bool]]):
    kwargs = {("min" if inclusive else "min_exclusive"): 3}
    column = cs.Decimal(**kwargs)  # type: ignore
    tbl = pdt.Table({"a": [1, 2, 3, 4, 5]})
    actual = evaluate_rules(tbl, column.validation_rules(tbl.a))
    assert actual == valid


@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
@pytest.mark.parametrize(
    ("inclusive", "valid"),
    [
        (True, {"max": [True, True, True, False, False]}),
        (False, {"max_exclusive": [True, True, False, False, False]}),
    ],
)
def test_validate_max(inclusive: bool, valid: dict[str, list[bool]]):
    kwargs = {("max" if inclusive else "max_exclusive"): decimal.Decimal(3)}
    column = cs.Decimal(**kwargs)  # type: ignore
    tbl = pdt.Table({"a": [1, 2, 3, 4, 5]})
    actual = evaluate_rules(tbl, column.validation_rules(tbl.a))
    assert actual == valid


@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
@pytest.mark.parametrize(
    ("min_inclusive", "max_inclusive", "valid"),
    [
        (
            True,
            True,
            {
                "min": [False, True, True, True, True],
                "max": [True, True, True, True, False],
            },
        ),
        (
            True,
            False,
            {
                "min": [False, True, True, True, True],
                "max_exclusive": [True, True, True, False, False],
            },
        ),
        (
            False,
            True,
            {
                "min_exclusive": [False, False, True, True, True],
                "max": [True, True, True, True, False],
            },
        ),
        (
            False,
            False,
            {
                "min_exclusive": [False, False, True, True, True],
                "max_exclusive": [True, True, True, False, False],
            },
        ),
    ],
)
def test_validate_range(
    min_inclusive: bool,
    max_inclusive: bool,
    valid: dict[str, list[bool]],
):
    kwargs = {
        ("min" if min_inclusive else "min_exclusive"): decimal.Decimal(2),
        ("max" if max_inclusive else "max_exclusive"): decimal.Decimal(4),
    }
    column = cs.Decimal(**kwargs)  # type: ignore
    tbl = pdt.Table({"a": [1, 2, 3, 4, 5]})
    actual = evaluate_rules(tbl, column.validation_rules(tbl.a))
    assert actual == valid



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

import pydiverse.colspec as cs
from pydiverse.colspec.optional_dependency import dy, pl

# -------------------------------------- SCHEMA -------------------------------------- #


class DepartmentColSpec(cs.ColSpec):
    department_id = cs.Int64(primary_key=True)


class ManagerColSpec(cs.ColSpec):
    department_id = cs.Int64(primary_key=True)
    name = cs.String(nullable=False)


class EmployeeColSpec(cs.ColSpec):
    department_id = cs.Int64(primary_key=True)
    employee_number = cs.Int64(primary_key=True)
    name = cs.String(nullable=False)


# ------------------------------------- FIXTURES ------------------------------------- #


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.fixture()
def departments() -> dy.LazyFrame[DepartmentColSpec]:
    return DepartmentColSpec.cast_polars(pl.LazyFrame({"department_id": [1, 2]}))


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.fixture()
def managers() -> dy.LazyFrame[ManagerColSpec]:
    return ManagerColSpec.cast_polars(
        pl.LazyFrame({"department_id": [1], "name": ["Donald Duck"]})
    )


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.fixture()
def employees() -> dy.LazyFrame[EmployeeColSpec]:
    return EmployeeColSpec.cast_polars(
        pl.LazyFrame(
            {
                "department_id": [2, 2, 2],
                "employee_number": [101, 102, 103],
                "name": ["Huey", "Dewey", "Louie"],
            }
        )
    )


# ------------------------------------------------------------------------------------ #
#                                         TESTS                                        #
# ------------------------------------------------------------------------------------ #


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_one_to_one(
    departments: dy.LazyFrame[DepartmentColSpec],
    managers: dy.LazyFrame[ManagerColSpec],
):
    actual = dy.filter_relationship_one_to_one(
        departments, managers, on="department_id"
    )
    assert actual.select("department_id").collect().to_series().to_list() == [1]


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_one_to_at_least_one(
    departments: dy.LazyFrame[DepartmentColSpec],
    employees: dy.LazyFrame[EmployeeColSpec],
):
    actual = dy.filter_relationship_one_to_at_least_one(
        departments, employees, on="department_id"
    )
    assert actual.select("department_id").collect().to_series().to_list() == [2]



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

import pydiverse.colspec as cs
from pydiverse.colspec.optional_dependency import dy, pl, validation_mask
from pydiverse.colspec.testing import create_colspec


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize("inner", [cs.Int64(), cs.Integer()])
def test_integer_list(inner: dy.Column):
    spec = create_colspec("test", {"a": cs.List(inner)})
    assert spec.is_valid_polars(pl.DataFrame({"a": [[1], [2], [3]]}))


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_invalid_inner_type():
    spec = create_colspec("test", {"a": cs.List(cs.Int64())})
    assert not spec.is_valid_polars(pl.DataFrame({"a": [["1"], ["2"], ["3"]]}))


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize(
    ("column", "dtype", "is_valid_polars"),
    [
        (
            cs.List(cs.Int64()),
            pl.List(pl.Int64()),
            True,
        ),
        (
            cs.List(cs.String()),
            pl.List(pl.Int64()),
            False,
        ),
        (
            cs.List(cs.String()),
            cs.List(cs.String()),
            False,
        ),
        (
            cs.List(cs.String()),
            cs.String(),
            False,
        ),
        (
            cs.List(cs.String()),
            pl.String(),
            False,
        ),
    ],
)
def test_validate_dtype(column: dy.Column, dtype: pl.DataType, is_valid_polars: bool):
    assert column.validate_dtype_polars(dtype) == is_valid_polars


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_nested_lists():
    spec = create_colspec("test", {"a": cs.List(cs.List(cs.Int64()))})
    assert spec.is_valid_polars(pl.DataFrame({"a": [[[1]], [[2]], [[3]]]}))


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_list_with_pk():
    spec = create_colspec(
        "test",
        {"a": cs.List(cs.String(), primary_key=True)},
    )
    df = pl.DataFrame({"a": [["ab"], ["a", "ab"], [None], ["a", "b"], ["a", "b"]]})
    _, failures = spec.filter_polars(df)
    assert validation_mask(df, failures).to_list() == [True, True, True, False, False]
    assert failures.counts() == {"primary_key": 2}


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_list_with_rules():
    spec = create_colspec(
        "test", {"a": cs.List(cs.String(min_length=2, nullable=False))}
    )
    df = pl.DataFrame({"a": [["ab"], ["a"], [None]]})
    _, failures = spec.filter_polars(df)
    assert validation_mask(df, failures).to_list() == [True, False, False]
    assert failures.counts() == {"a|inner_nullability": 1, "a|inner_min_length": 1}


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_nested_list_with_rules():
    spec = create_colspec(
        "test", {"a": cs.List(cs.List(cs.String(min_length=2, nullable=False)))}
    )
    df = pl.DataFrame({"a": [[["ab"]], [["a"]], [[None]]]})
    _, failures = spec.filter_polars(df)
    # NOTE: `validation_mask` currently fails for multiply nested lists
    assert failures.invalid().to_dict(as_series=False) == {"a": [[["a"]], [[None]]]}
    assert failures.counts() == {
        "a|inner_inner_nullability": 1,
        "a|inner_inner_min_length": 1,
    }


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_list_length_rules():
    spec = create_colspec(
        "test",
        {
            "a": cs.List(
                cs.Integer(nullable=False),
                min_length=2,
                max_length=5,
                nullable=True,
            )
        },
    )
    df = pl.DataFrame({"a": [[31, 12], [-1], [None], None, [1, 2, 3, 4, 23, 1]]})
    _, failures = spec.filter_polars(df)
    assert validation_mask(df, failures).to_list() == [True, False, False, True, False]


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_outer_inner_nullability():
    spec = create_colspec(
        "test",
        {
            "nullable": cs.List(
                inner=cs.Integer(nullable=False),
                nullable=True,
            )
        },
    )
    df = pl.DataFrame({"nullable": [None, None]})
    spec.validate_polars(df, cast=True)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_inner_primary_key():
    spec = create_colspec("test", {"a": cs.List(cs.Integer(primary_key=True))})
    df = pl.DataFrame({"a": [[1, 2, 3], [1, 1, 2], [1, 1], [1, 4]]})
    _, failure = spec.filter_polars(df)
    assert failure.counts() == {"a|primary_key": 2}
    assert validation_mask(df, failure).to_list() == [True, False, False, True]


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize(
    ("inner_primary_key", "second_primary_key", "failure_count", "mask"),
    [
        (False, True, 2, [False, False, True, True, True, True, True]),
        (True, True, 1, [True, False, True, True, True, True, True]),
        (False, False, 4, [False, False, True, False, False, True, True]),
        (True, False, 1, [True, False, True, True, True, True, True]),
    ],
)
def test_inner_primary_key_struct(
    inner_primary_key: bool,
    second_primary_key: bool,
    failure_count: int,
    mask: list[bool],
):
    spec = create_colspec(
        "test",
        {
            "a": cs.List(
                cs.Struct(
                    {
                        "pk1": cs.Integer(primary_key=True),
                        "pk2": cs.Integer(primary_key=second_primary_key),
                        "other": cs.Integer(),
                    },
                    primary_key=inner_primary_key,
                )
            )
        },
    )
    df = pl.DataFrame(
        {
            "a": [
                [{"pk1": 1, "pk2": 1, "other": 1}, {"pk1": 1, "pk2": 1, "other": 2}],
                [{"pk1": 1, "pk2": 1, "other": 1}, {"pk1": 1, "pk2": 1, "other": 1}],
                [{"pk1": 1, "pk2": 1, "other": 1}],
                [{"pk1": 1, "pk2": 1, "other": 1}, {"pk1": 1, "pk2": 2, "other": 1}],
                [{"pk1": 1, "pk2": 1, "other": 1}, {"pk1": 1, "pk2": 2, "other": 2}],
                [{"pk1": 1, "pk2": 1, "other": 1}, {"pk1": 2, "pk2": 2, "other": 2}],
                [],
            ]
        }
    )
    _, failure = spec.filter_polars(df)
    assert failure.counts() == {"a|primary_key": failure_count}
    assert validation_mask(df, failure).to_list() == mask



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from ._filter import Filter, FilterPolars, filter, filter_polars  # noqa: A004
from ._rule import GroupRule, GroupRulePolars, Rule, RulePolars, rule, rule_polars
from .collection import Collection
from .colspec import ColSpec
from .columns import (
    Any,
    Bool,
    Column,
    Date,
    Datetime,
    Decimal,
    Duration,
    Enum,
    Float,
    Float32,
    Float64,
    Int8,
    Int16,
    Int32,
    Int64,
    Integer,
    List,
    String,
    Struct,
    Time,
    UInt8,
    UInt16,
    UInt32,
    UInt64,
)
from .version import __version__

__all__ = [
    "__version__",
    "ColSpec",
    "Collection",
    "filter",
    "Filter",
    "filter_polars",
    "FilterPolars",
    "rule",
    "rule_polars",
    "Rule",
    "RulePolars",
    "GroupRule",
    "GroupRulePolars",
    "Column",
    "Any",
    "Bool",
    "Date",
    "Datetime",
    "Decimal",
    "Duration",
    "Enum",
    "Time",
    "Float",
    "Float32",
    "Float64",
    "Int8",
    "Int16",
    "Int32",
    "Int64",
    "Integer",
    "UInt8",
    "UInt16",
    "UInt32",
    "UInt64",
    "String",
    "List",
    "Struct",
]



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

import pydiverse.colspec as cs
from pydiverse.colspec.columns._base import Column
from pydiverse.colspec.optional_dependency import dy, pl
from pydiverse.colspec.testing.factory import create_colspec


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_simple_struct():
    schema = create_colspec(
        "test", {"s": cs.Struct({"a": cs.Integer(), "b": cs.String()})}
    )
    assert schema.is_valid_polars(
        pl.DataFrame({"s": [{"a": 1, "b": "foo"}, {"a": 2, "b": "foo"}]})
    )


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize(
    ("column", "dtype", "is_valid"),
    [
        (
            cs.Struct({"a": cs.Int64(), "b": cs.String()}),
            pl.Struct({"a": pl.Int64(), "b": pl.String()}),
            True,
        ),
        (
            cs.Struct({"b": cs.String(), "a": cs.Int64()}),
            pl.Struct({"a": pl.Int64(), "b": pl.String()}),
            True,
        ),
        (
            cs.Struct({"a": cs.Int64(), "b": cs.String(), "c": cs.String()}),
            pl.Struct({"a": pl.Int64(), "b": pl.String()}),
            False,
        ),
        (
            cs.Struct({"a": cs.String(), "b": cs.Int64()}),
            pl.Struct({"a": pl.Int64(), "b": pl.String()}),
            False,
        ),
        (
            cs.Struct({"a": cs.String(), "b": cs.Int64()}),
            pl.Struct({"a": pl.Int64(), "b": pl.String()}),
            False,
        ),
        (
            cs.Struct({"a": cs.String(), "b": cs.Int64()}),
            pl.Struct({"a": pl.Int64(), "b": pl.String()}),
            False,
        ),
        (
            cs.Struct({"a": cs.Int64(), "b": cs.String()}),
            pl.Struct({"a": pl.Int64(), "c": pl.String()}),
            False,
        ),
        (
            cs.Struct({"a": cs.String(), "b": cs.Int64()}),
            cs.Struct({"a": cs.String(), "b": cs.Int64()}),
            False,
        ),
        (
            cs.Struct({"a": cs.String(), "b": cs.Int64()}),
            cs.String(),
            False,
        ),
        (
            cs.Struct({"a": cs.String(), "b": cs.Int64()}),
            pl.String(),
            False,
        ),
    ],
)
def test_validate_dtype(column: Column, dtype: pl.DataType, is_valid: bool):
    assert column.validate_dtype_polars(dtype) == is_valid


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_invalid_inner_type():
    schema = create_colspec("test", {"a": cs.Struct({"a": cs.Int64()})})
    assert not schema.is_valid_polars(pl.DataFrame({"a": [{"a": "1"}, {"a": "2"}]}))


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_nested_structs():
    schema = create_colspec(
        "test",
        {
            "s1": cs.Struct(
                {
                    "s2": cs.Struct({"a": cs.Integer(), "b": cs.String()}),
                    "c": cs.String(),
                }
            )
        },
    )
    assert schema.is_valid_polars(
        pl.DataFrame({"s1": [{"s2": {"a": 1, "b": "foo"}, "c": "bar"}]})
    )


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_struct_with_pk():
    schema = create_colspec(
        "test",
        {"s": cs.Struct({"a": cs.String(), "b": cs.Integer()}, primary_key=True)},
    )
    df = pl.DataFrame(
        {"s": [{"a": "foo", "b": 1}, {"a": "bar", "b": 1}, {"a": "bar", "b": 1}]}
    )
    _, failures = schema.filter_polars(df)
    assert failures.invalid().to_dict(as_series=False) == {
        "s": [{"a": "bar", "b": 1}, {"a": "bar", "b": 1}]
    }
    assert failures.counts() == {"primary_key": 2}


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_struct_with_rules():
    schema = create_colspec(
        "test", {"s": cs.Struct({"a": cs.String(min_length=2, nullable=False)})}
    )
    df = pl.DataFrame({"s": [{"a": "ab"}, {"a": "a"}, {"a": None}]})
    _, failures = schema.filter_polars(df)
    assert failures.invalid().to_dict(as_series=False) == {
        "s": [{"a": "a"}, {"a": None}]
    }
    assert failures.counts() == {"s|inner_a_nullability": 1, "s|inner_a_min_length": 1}


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_nested_struct_with_rules():
    schema = create_colspec(
        "test",
        {
            "s1": cs.Struct(
                {"s2": cs.Struct({"a": cs.String(min_length=2, nullable=False)})}
            )
        },
    )
    df = pl.DataFrame(
        {"s1": [{"s2": {"a": "ab"}}, {"s2": {"a": "a"}}, {"s2": {"a": None}}]}
    )
    _, failures = schema.filter_polars(df)
    assert failures.invalid().to_dict(as_series=False) == {
        "s1": [{"s2": {"a": "a"}}, {"s2": {"a": None}}]
    }
    assert failures.counts() == {
        "s1|inner_s2_inner_a_nullability": 1,
        "s1|inner_s2_inner_a_min_length": 1,
    }


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_outer_inner_nullability():
    schema = create_colspec(
        "test",
        {
            "nullable": cs.Struct(
                inner={
                    "not_nullable1": cs.Integer(nullable=False),
                    "not_nullable2": cs.Integer(nullable=False),
                },
                nullable=True,
            )
        },
    )
    df = pl.DataFrame({"nullable": [None, None]})

    schema.validate_polars(df, cast=True)



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import pydiverse.colspec as cs

COLUMN_TYPES: list[type[cs.Column]] = [
    cs.Bool,
    cs.Date,
    cs.Datetime,
    cs.Time,
    cs.Decimal,
    cs.Duration,
    cs.Float32,
    cs.Float64,
    cs.Int8,
    cs.Int16,
    cs.Int32,
    cs.Int64,
    cs.UInt8,
    cs.UInt16,
    cs.UInt32,
    cs.UInt64,
    cs.String,
]
INTEGER_COLUMN_TYPES: list[type[cs.Column]] = [
    cs.Integer,
    cs.Int8,
    cs.Int16,
    cs.Int32,
    cs.Int64,
    cs.UInt8,
    cs.UInt16,
    cs.UInt32,
    cs.UInt64,
]
FLOAT_COLUMN_TYPES: list[type[cs.Column]] = [
    cs.Float,
    cs.Float32,
    cs.Float64,
]

SUPERTYPE_COLUMN_TYPES: list[type[cs.Column]] = [
    cs.Float,
    cs.Integer,
]

ALL_COLUMN_TYPES: list[type[cs.Column]] = (
    [cs.Any] + COLUMN_TYPES + SUPERTYPE_COLUMN_TYPES
)



# Copyright (c) QuantCo and pydiverse contributors 2023-2025
# SPDX-License-Identifier: BSD-3-Clause
import sys
from abc import abstractmethod
from collections.abc import Callable

import pydiverse.common as pdc

from ..optional_dependency import ColExpr
from ._base import Column
from ._mixins import OrdinalMixin
from ._utils import classproperty


class _BaseFloat(OrdinalMixin[float], Column):
    def __init__(
        self,
        *,
        nullable: bool = True,
        primary_key: bool = False,
        min: float | None = None,  # noqa: A002
        min_exclusive: float | None = None,
        max: float | None = None,  # noqa: A002
        max_exclusive: float | None = None,
        check: Callable[[ColExpr], ColExpr] | None = None,
        alias: str | None = None,
    ):
        """
        Args:
            nullable: Whether this column may contain null values.
            primary_key: Whether this column is part of the primary key of the schema.
                If ``True``, ``nullable`` is automatically set to ``False``.
            min: The minimum value for floats in this column (inclusive).
            min_exclusive: Like ``min`` but exclusive. May not be specified if ``min``
                is specified and vice versa.
            max: The maximum value for floats in this column (inclusive).
            max_exclusive: Like ``max`` but exclusive. May not be specified if ``max``
                is specified and vice versa.
            check: A custom check to run for this column. Must return a non-aggregated
                boolean expression.
            alias: An overwrite for this column's name which allows for using a column
                name that is not a valid Python identifier. Especially note that setting
                this option does _not_ allow to refer to the column with two different
                names, the specified alias is the only valid name.
        """
        if min is not None and min < self.min_value:
            raise ValueError("Minimum value is too small for the data type.")
        if max is not None and max > self.max_value:
            raise ValueError("Maximum value is too big for the data type.")

        super().__init__(
            nullable=nullable,
            primary_key=primary_key,
            min=min,
            min_exclusive=min_exclusive,
            max=max,
            max_exclusive=max_exclusive,
            check=check,
            alias=alias,
        )

    @classproperty
    @abstractmethod
    def max_value(self) -> float:
        """Maximum value of the column's type."""

    @classproperty
    @abstractmethod
    def min_value(self) -> float:
        """Minimum value of the column's type."""


# ------------------------------------------------------------------------------------ #


class Float(_BaseFloat):
    """A column of floating-point numbers."""

    def dtype(self) -> pdc.Float:
        return pdc.Float()

    @classproperty
    def max_value(self) -> float:
        return sys.float_info.max

    @classproperty
    def min_value(self) -> float:
        return -sys.float_info.max


class Float32(_BaseFloat):
    """A column of 32-bit floating-point numbers."""

    def dtype(self) -> pdc.Float32:
        return pdc.Float32()

    @classproperty
    def max_value(self) -> float:
        return 3.4028234663852886e38  # float(np.finfo(np.float32).max)

    @classproperty
    def min_value(self) -> float:
        return -3.4028234663852886e38  # float(np.finfo(np.float32).min)


class Float64(_BaseFloat):
    """A column of 64-bit floating-point numbers."""

    def dtype(self) -> pdc.Float64:
        return pdc.Float64()

    @classproperty
    def max_value(self) -> float:
        return sys.float_info.max

    @classproperty
    def min_value(self) -> float:
        return -sys.float_info.max



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any

import pytest

import pydiverse.colspec as cs
from pydiverse.colspec.optional_dependency import C, dy, pdt, pl


class AnyColSpec(cs.ColSpec):
    a = cs.Any()


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
@pytest.mark.parametrize(
    "data",
    [{"a": [None]}, {"a": [True, None]}, {"a": ["foo"]}, {"a": [3.5]}],
)
def test_any_dtype_passes(data: dict[str, Any]):
    df = pl.DataFrame(data)
    assert AnyColSpec.is_valid_polars(df)
    tbl = pdt.Table(df)
    with pytest.raises(NotImplementedError, match="intentionally not implemented"):
        assert AnyColSpec.is_valid(tbl)



# Copyright (c) QuantCo and pydiverse contributors 2023-2025
# SPDX-License-Identifier: BSD-3-Clause

from collections import defaultdict
from collections.abc import Iterable

from pydiverse.common import Dtype


class ValidationError(Exception):
    """Error raised when :mod:`dataframely` validation encounters an issue."""

    def __init__(self, message: str):
        super().__init__()
        self.message = message

    def __str__(self) -> str:
        return self.message


class ColumnValidationError(ValidationError):
    """Validation error raised when columns mismatch."""

    def __init__(
        self,
        missing: Iterable[str] = tuple(),
        extra: Iterable[str] = tuple(),
        actual: Iterable[str] = tuple(),
    ):
        msg = []
        if missing:
            msg.append(f"Missing columns: {', '.join(missing)}")
        if extra:
            msg.append(f"Additional columns: {', '.join(extra)}")
        super().__init__(
            f"{len(missing)} columns are missing: {', '.join(missing)}; "
            f"found: {', '.join(actual)}"
            if actual
            else "; ".join(msg)
            if msg
            else "Column validation failed"
        )
        self.missing = missing
        self.extra = extra
        self.actual = actual

    def __str__(self) -> str:
        details = [
            f" - Missing columns: {', '.join(self.missing)}",
            f" - Extra columns: {', '.join(self.extra)}",
            f" - Actual columns: {', '.join(self.actual)}",
        ]
        return "\n".join([f"{self.message}:"] + details)


class DtypeValidationError(ValidationError):
    """Validation error raised when column dtypes are wrong."""

    def __init__(self, errors: dict[str, tuple[Dtype, Dtype]]):
        super().__init__(f"{len(errors)} columns have an invalid dtype")
        self.errors = errors

    def __str__(self) -> str:
        details = [
            f" - '{col}': got dtype '{actual}' but expected '{expected}'"
            for col, (actual, expected) in self.errors.items()
        ]
        return "\n".join([f"{self.message}:"] + details)


class RuleValidationError(ValidationError):
    """Complex validation error raised when rule validation fails."""

    def __init__(self, errors: dict[str, int]):
        super().__init__(f"{len(errors)} rules failed validation")

        # Split into schema errors and column errors
        schema_errors: dict[str, int] = {}
        column_errors: dict[str, dict[str, int]] = defaultdict(dict)
        for name, count in sorted(errors.items()):
            if "|" in name:
                column, rule = name.split("|", maxsplit=1)
                column_errors[column][rule] = count
            else:
                schema_errors[name] = count

        self.schema_errors = schema_errors
        self.column_errors = column_errors

    def __str__(self) -> str:
        schema_details = [
            f" - '{name}' failed validation for {count:,} rows"
            for name, count in self.schema_errors.items()
        ]
        column_details = [
            msg
            for column, errors in self.column_errors.items()
            for msg in (
                [f" * Column '{column}' failed validation for {len(errors)} rules:"]
                + [
                    f"   - '{name}' failed for {count:,} rows"
                    for name, count in errors.items()
                ]
            )
        ]
        return "\n".join([f"{self.message}:"] + schema_details + column_details)


class MemberValidationError(ValidationError):
    """Validation error raised when multiple members of a collection fail validation."""

    def __init__(self, errors: dict[str, ValidationError]):
        super().__init__(f"{len(errors)} members failed validation")
        self.errors = errors

    def __str__(self):
        details = [
            f" > Member '{name}' failed validation:\n"
            + "\n".join("   " + line for line in str(error).split("\n"))
            for name, error in self.errors.items()
        ]
        return "\n".join([f"{self.message}:"] + details)


class ImplementationError(Exception):
    """Error raised when a schema is implemented incorrectly."""


class AnnotationImplementationErrorDetail(ImplementationError):
    def __init__(self, message: str, _type: type):
        self._type = _type
        super().__init__(message)


class AnnotationImplementationError(ImplementationError):
    """Error raised when the annotations of a collection are invalid."""

    def __init__(self, attr: str, kls: type):
        message = (
            "Annotations of a 'dy.Collection' may only be an (optional) "
            f"'ColSpec', but \"{attr}\" has type '{kls}'."
        )
        super().__init__(message)


class RuleImplementationError(ImplementationError):
    """Error raised when a rule is implemented incorrectly."""

    def __init__(self, name: str, return_dtype: Dtype, is_group_rule: bool):
        if is_group_rule:
            details = (
                " When implementing a group rule (i.e. when using the `group_by` "
                "parameter), make sure to use an aggregation function such as `.any()`,"
                " `.all()`, and others to reduce an expression evaluated on multiple "
                "rows in the same group to a single boolean value for the group."
            )
        else:
            details = ""

        message = (
            f"Validation rule '{name}' has not been implemented correctly. It "
            f"returns dtype '{return_dtype}' but it must return a boolean value."
            + details
        )
        super().__init__(message)



# Copyright (c) QuantCo and pydiverse contributors 2023-2025
# SPDX-License-Identifier: BSD-3-Clause

import datetime as dt
from abc import ABC, ABCMeta, abstractmethod
from collections.abc import Callable

from pydiverse.colspec.optional_dependency import dy
from pydiverse.common import Dtype

from ..optional_dependency import ColExpr, Generator, PolarsDataType, pa, pl, sa
from ._utils import pydiverse_type_opinions

EPOCH_DATETIME = dt.datetime(1970, 1, 1)
SECONDS_PER_DAY = 86400

# ------------------------------------------------------------------------------------ #
#                                        COLUMNS                                       #
# ------------------------------------------------------------------------------------ #


class ColumnMeta(ABCMeta):
    def __new__(cls, clsname, bases, attribs):
        # change bases (only ABC is a real base)
        bases = tuple([base for base in bases if not issubclass(base, ColExpr)])
        return super().__new__(cls, clsname, bases, attribs)


class Column(ABC, ColExpr, metaclass=ColumnMeta):
    """Abstract base class for data frame column definitions.

    This class is merely supposed to be used in :class:`~colspec.ColSpec`
    definitions.
    """

    def __init__(
        self,
        *,
        nullable: bool = True,
        primary_key: bool = False,
        check: Callable[[ColExpr], ColExpr] | None = None,
        alias: str | None = None,
    ):
        """
        Args:
            nullable: Whether this column may contain null values.
            primary_key: Whether this column is part of the primary key of the schema.
                If ``True``, ``nullable`` is automatically set to ``False``.
            check: A custom check to run for this column. Must return a non-aggregated
                boolean expression.
            alias: An overwrite for this column's name which allows for using a column
                name that is not a valid Python identifier. Especially note that setting
                this option does _not_ allow to refer to the column with two different
                names, the specified alias is the only valid name. If unset, colspec
                internally sets the alias to the column's name in the parent schema.
        """
        self.nullable = nullable and not primary_key
        self.primary_key = primary_key
        self.check = check
        self.alias = alias

    # ------------------------------------- DTYPE ------------------------------------ #

    @abstractmethod
    def dtype(self) -> Dtype:
        """The common dtype of this column definition.

        Returns a pydiverse.common.Dtype instance that represents the abstract data type
        of this column. This type can then be mapped to specific backend types (SQL,
        Polars, etc.) through the respective conversion methods.

        Returns:
            A pydiverse.common.Dtype instance representing the column's data type.
        """

    # ---------------------------------- VALIDATION ---------------------------------- #

    def validation_rules(self, expr: ColExpr) -> dict[str, ColExpr]:
        """A set of rules evaluating whether a data frame column satisfies the column's
        constraints.

        Args:
            expr: An expression referencing the column of the data frame, i.e. an
                expression created by calling :meth:`polars.col`.

        Returns:
            A mapping from validation rule names to expressions that provide exactly
            one boolean value per column item indicating whether validation with respect
            to the rule is successful. A value of ``False`` indicates invalid data, i.e.
            unsuccessful validation.
        """
        result = {}
        if not self.nullable:
            result["nullability"] = expr.is_not_null()
        if self.check is not None:
            result["check"] = self.check(expr)
        return result

    # -------------------------------- POLARS VALIDATION ----------------------------- #

    def to_dataframely(self):
        """Convert this column to its dataframely equivalent.

        Returns:
            A dataframely.Column instance with the same properties as this column.
        """

        def convert(value):
            if isinstance(value, Column):
                return value.to_dataframely()
            if isinstance(value, dict):
                return {k: convert(v) for k, v in value.items()}
            if isinstance(value, list):
                return [convert(v) for v in value]
            if isinstance(value, tuple):
                return tuple(convert(v) for v in value)
            return value

        # Get all non-private attributes
        attrs = {
            k: convert(v) for k, v in self.__dict__.items() if not k.startswith("_")
        }
        return getattr(dy, self.__class__.__name__)(**attrs)

    def validate_dtype_polars(self, dtype: PolarsDataType) -> bool:
        """Validate if the :mod:`polars` data type satisfies the column definition.

        This function requires dataframely to be installed since this is used as colspec
        implementation for polars.
        Args:
            dtype: The dtype to validate.

        Returns:
            Whether the dtype is valid.
        """
        return self.to_dataframely().validate_dtype(dtype)

    # -------------------------------- POLARS SAMPLING ------------------------------- #

    def sample_polars(self, generator: Generator, n: int = 1) -> pl.Series:
        """Sample random elements adhering to the constraints of this column.

        This function requires dataframely to be installed since this is used as colspec
        implementation for polars.
        Args:
            generator: The generator to use for sampling elements.
            n: The number of elements to sample.

        Returns:
            A series with the predefined number of elements. All elements are guaranteed
            to adhere to the column's constraints.

        Raises:
            ValueError: If this column has a custom check. In this case, random values
                cannot be guaranteed to adhere to the column's constraints while
                providing any guarantees on the computational complexity.
        """
        return self.to_dataframely().sample(generator, n)

    # -------------------------------------- SQL ------------------------------------- #

    def sqlalchemy_column(self, name: str, dialect: sa.Dialect) -> sa.Column:
        """Obtain the SQL column specification of this column definition.

        Args:
            name: The name of the column.
            dialect: The SQL dialect for which to generate the column specification.

        Returns:
            The column as specified in :mod:`sqlalchemy`.
        """
        _ = dialect  # may be used in the future
        return sa.Column(
            name,
            self.dtype().to_sql(),
            nullable=self.nullable,
            primary_key=self.primary_key,
            autoincrement=False,
        )

    # ------------------------------------ PYARROW ----------------------------------- #

    def pyarrow_field(self, name: str) -> pa.Field:
        """Obtain the pyarrow field of this column definition.

        Args:
            name: The name of the column.

        Returns:
            The :mod:`pyarrow` field definition.
        """
        try:
            return pa.field(name, self.dtype().to_arrow(), nullable=self.nullable)
        except NotImplementedError:
            return pa.field(
                name,
                pydiverse_type_opinions(self.to_dataframely().pyarrow_dtype),
                nullable=self.nullable,
            )

    # -------------------------------- DUNDER METHODS -------------------------------- #

    def __str__(self) -> str:
        return self.__class__.__name__.lower()



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import types
from typing import Generic, TypeVar

try:
    import numpy as np
except ImportError:
    np = None


try:
    import polars as pl
    from polars.datatypes import DataTypeClass

    PolarsDataType = pl.DataType | DataTypeClass
    import polars.exceptions as plexc
    from polars.datatypes.group import FLOAT_DTYPES, INTEGER_DTYPES
    from polars.testing import assert_frame_equal
except ImportError:

    class DummyClass:
        def __init__(self, *args, **kwargs):
            pass

    DataTypeClass = None
    PolarsDataType = None
    FLOAT_DTYPES, INTEGER_DTYPES = [], []
    assert_frame_equal = None
    plexc = None
    # Create a new module with the given name.
    pl = types.ModuleType("polars")
    pl.DataFrame = DummyClass
    pl.LazyFrame = DummyClass
    pl.DataType = None
    pl.Series = None
    pl.Schema = None
    pl.Expr = None
    for _type in [
        "Int8",
        "Int16",
        "Int32",
        "Int64",
        "UInt8",
        "UInt16",
        "UInt32",
        "UInt64",
        "Float32",
        "Float64",
        "Boolean",
        "Utf8",
        "Decimal",
        "Enum",
        "Struct",
        "List",
        "Date",
        "Datetime",
        "Time",
        "Duration",
        "String",
        "Null",
    ]:
        setattr(pl, _type, DummyClass)


try:
    # colspec has optional dependency to dataframely
    import dataframely as dy
    from dataframely._polars import FrameType
    from dataframely.random import Generator
    from dataframely.testing import validation_mask
except ImportError:

    class Generator:
        pass

    T = TypeVar("T")

    class DyDataFrame(Generic[T]):
        pass

    class DyDummyClass:
        pass

    FrameType = None
    validation_mask = None
    dy = types.ModuleType("dataframely")
    dy.DataFrame = DyDataFrame
    dy.LazyFrame = DyDataFrame
    dy.FailureInfo = None
    dy.Column = None
    dy.Collection = None
    for _type in [
        "Int8",
        "Int16",
        "Int32",
        "Int64",
        "UInt8",
        "UInt16",
        "UInt32",
        "UInt64",
        "Float32",
        "Float64",
        "Bool",
        "String",
        "Decimal",
        "Enum",
        "Struct",
        "List",
        "Date",
        "Datetime",
        "Time",
        "Duration",
        "Float",
        "Integer",
    ]:
        setattr(dy, _type, DyDummyClass)


try:
    # colspec has optional dependency to pydiverse.transform
    import pydiverse.transform as pdt
    from pydiverse.transform import C, ColExpr, verb
except ImportError:

    def verb(func):
        """A no-op decorator for functions that are intended to be used as verbs."""
        return func

    class Table:
        pass

    class ColExpr:
        pass

    # Create a new module with the given name.
    pdt = types.ModuleType("pydiverse.transform")
    pdt.Table = Table
    C = None
    # TODO: add members that break if pdt is not there


try:
    # colspec has optional dependency to pydiverse.pipedag
    import pydiverse.pipedag as dag
except ImportError:

    class Table:
        pass

    # Create a new module with the given name.
    dag = types.ModuleType("pydiverse.pipedag")
    dag.Table = Table


try:
    # colspec has optional dependency to pyarrow
    import pyarrow as pa
except ImportError:
    pa = types.ModuleType("pyarrow")
    pa.Schema = None
    pa.Field = None
    pa.DataType = None


try:
    # colspec has optional dependency to sqlalchemy
    import pyodbc
    import sqlalchemy as sa
except ImportError:
    sa = types.ModuleType("sqlalchemy")
    sa.Dialect = None
    sa.Column = None
    sa.Table = None
    sa.Alias = None
    pyodbc = None



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

import dataclasses

import pytest

import pydiverse.colspec as cs
import pydiverse.colspec.collection
from pydiverse.colspec.exc import MemberValidationError
from pydiverse.colspec.optional_dependency import assert_frame_equal, dy, pl

pytestmark = pytest.mark.skipif(
    dy.Column is None, reason="dataframely is required for this test"
)

# ------------------------------------------------------------------------------------ #
#                                        SCHEMA                                        #
# ------------------------------------------------------------------------------------ #


class MyFirstColSpec(cs.ColSpec):
    a = cs.Integer(primary_key=True)
    b = cs.Integer()


class MySecondColSpec(cs.ColSpec):
    a = cs.Integer(primary_key=True)
    b = cs.Integer(min=1)


class MyCollection(pydiverse.colspec.collection.Collection):
    first: MyFirstColSpec
    second: MySecondColSpec

    @cs.filter_polars()
    def equal_primary_keys(self) -> pl.LazyFrame:
        return self.first.join(self.second, on=self.common_primary_keys())

    @cs.filter_polars()
    def first_b_greater_second_b(self) -> pl.LazyFrame:
        return self.first.join(
            self.second, on=self.common_primary_keys(), how="full", coalesce=True
        ).filter((pl.col("b") > pl.col("b_right")).fill_null(True))


@dataclasses.dataclass
class SimpleCollection(pydiverse.colspec.collection.Collection):
    first: MyFirstColSpec
    second: MySecondColSpec


# ------------------------------------------------------------------------------------ #
#                                         TESTS                                        #
# ------------------------------------------------------------------------------------ #


@pytest.fixture()
def data_without_filter_without_rule_violation() -> tuple[pl.LazyFrame, pl.LazyFrame]:
    first = pl.LazyFrame({"a": [1, 2, 3], "b": [1, 2, 3]})
    second = pl.LazyFrame({"a": [1, 2, 3], "b": [1, 2, 3]})
    return first, second


@pytest.fixture()
def data_without_filter_with_rule_violation() -> tuple[pl.LazyFrame, pl.LazyFrame]:
    first = pl.LazyFrame({"a": [1, 2, 1], "b": [1, 2, 3]})
    second = pl.LazyFrame({"a": [1, 2, 3], "b": [0, 1, 2]})
    return first, second


@pytest.fixture()
def data_with_filter_without_rule_violation() -> tuple[pl.LazyFrame, pl.LazyFrame]:
    first = pl.LazyFrame({"a": [1, 2, 3], "b": [1, 1, 3]})
    second = pl.LazyFrame({"a": [2, 3, 4, 5], "b": [1, 2, 3, 4]})
    return first, second


@pytest.fixture()
def data_with_filter_with_rule_violation() -> tuple[pl.LazyFrame, pl.LazyFrame]:
    first = pl.LazyFrame({"a": [1, 2, 3], "b": [1, 2, 3]})
    second = pl.LazyFrame({"a": [2, 3, 4, 5], "b": [0, 1, 2, 3]})
    return first, second


# -------------------------------------- FILTER -------------------------------------- #


def test_filter_without_filter_without_rule_violation(
    data_without_filter_without_rule_violation: tuple[pl.LazyFrame, pl.LazyFrame],
):
    out, failure = SimpleCollection(
        data_without_filter_without_rule_violation[0],
        data_without_filter_without_rule_violation[1],
    ).filter_polars()

    assert isinstance(out, SimpleCollection)
    assert_frame_equal(out.first, data_without_filter_without_rule_violation[0])
    assert_frame_equal(out.second, data_without_filter_without_rule_violation[1])
    assert len(failure["first"]) == 0
    assert len(failure["second"]) == 0


def test_filter_without_filter_with_rule_violation(
    data_without_filter_with_rule_violation: tuple[pl.LazyFrame, pl.LazyFrame],
):
    out, failure = SimpleCollection.filter_polars_data(
        {
            "first": data_without_filter_with_rule_violation[0],
            "second": data_without_filter_with_rule_violation[1],
        }
    )

    assert isinstance(out, SimpleCollection)
    assert len(out.first.collect()) == 1
    assert len(out.second.collect()) == 2
    assert failure["first"].counts() == {"primary_key": 2}
    assert failure["second"].counts() == {"b|min": 1}


def test_filter_with_filter_without_rule_violation(
    data_with_filter_without_rule_violation: tuple[pl.LazyFrame, pl.LazyFrame],
):
    out, failure = MyCollection.filter_polars_data(
        {
            "first": data_with_filter_without_rule_violation[0],
            "second": data_with_filter_without_rule_violation[1],
        }
    )

    assert isinstance(out, MyCollection)
    assert_frame_equal(out.first, pl.LazyFrame({"a": [3], "b": [3]}))
    assert_frame_equal(out.second, pl.LazyFrame({"a": [3], "b": [2]}))
    assert failure["first"].counts() == {
        "equal_primary_keys": 1,
        "first_b_greater_second_b": 1,
    }
    assert failure["second"].counts() == {
        "equal_primary_keys": 2,
        "first_b_greater_second_b": 1,
    }


def test_filter_with_filter_with_rule_violation(
    data_with_filter_with_rule_violation: tuple[pl.LazyFrame, pl.LazyFrame],
):
    out, failure = MyCollection.filter_polars_data(
        {
            "first": data_with_filter_with_rule_violation[0],
            "second": data_with_filter_with_rule_violation[1],
        }
    )

    assert isinstance(out, MyCollection)
    assert_frame_equal(out.first, pl.LazyFrame({"a": [3], "b": [3]}))
    assert_frame_equal(out.second, pl.LazyFrame({"a": [3], "b": [1]}))
    assert failure["first"].counts() == {"equal_primary_keys": 2}
    assert failure["second"].counts() == {"b|min": 1, "equal_primary_keys": 2}


# -------------------------------- VALIDATE WITH DATA -------------------------------- #


def test_validate_without_filter_without_rule_violation(
    data_without_filter_without_rule_violation: tuple[pl.LazyFrame, pl.LazyFrame],
):
    data = {
        "first": data_without_filter_without_rule_violation[0],
        "second": data_without_filter_without_rule_violation[1],
    }
    assert SimpleCollection.is_valid_polars_data(data)
    out = SimpleCollection.validate_polars_data(data)

    assert isinstance(out, SimpleCollection)
    assert_frame_equal(out.first, data_without_filter_without_rule_violation[0])
    assert_frame_equal(out.second, data_without_filter_without_rule_violation[1])


def test_validate_without_filter_with_rule_violation(
    data_without_filter_with_rule_violation: tuple[pl.LazyFrame, pl.LazyFrame],
):
    data = {
        "first": data_without_filter_with_rule_violation[0],
        "second": data_without_filter_with_rule_violation[1],
    }
    assert not SimpleCollection.is_valid_polars_data(data)

    with pytest.raises(
        MemberValidationError, match=r"2 members failed validation"
    ) as exc:
        SimpleCollection.validate_polars_data(data)

    exc.match(r"Member 'first' failed validation")
    exc.match(r"'primary_key' failed validation for 2 rows")
    exc.match(r"Member 'second' failed validation")
    exc.match(r"'min' failed for 1 rows")


def test_validate_with_filter_without_rule_violation(
    data_with_filter_without_rule_violation: tuple[pl.LazyFrame, pl.LazyFrame],
):
    data = {
        "first": data_with_filter_without_rule_violation[0],
        "second": data_with_filter_without_rule_violation[1],
    }
    assert not MyCollection.is_valid_polars_data(data)

    with pytest.raises(
        MemberValidationError, match=r"2 members failed validation"
    ) as exc:
        MyCollection.validate_polars_data(data)

    exc.match(r"Member 'first' failed validation")
    exc.match(r"'equal_primary_keys' failed validation for 1 rows")
    exc.match(r"'first_b_greater_second_b' failed validation for 1 rows")
    exc.match(r"Member 'second' failed validation")
    exc.match(r"'equal_primary_keys' failed validation for 2 rows")


def test_validate_with_filter_with_rule_violation(
    data_with_filter_with_rule_violation: tuple[pl.LazyFrame, pl.LazyFrame],
):
    data = {
        "first": data_with_filter_with_rule_violation[0],
        "second": data_with_filter_with_rule_violation[1],
    }
    assert not MyCollection.is_valid_polars_data(data)

    with pytest.raises(
        MemberValidationError, match=r"2 members failed validation"
    ) as exc:
        MyCollection.validate_polars_data(data)

    exc.match(r"Member 'first' failed validation")
    exc.match(r"'equal_primary_keys' failed validation for 2 rows")
    exc.match(r"Member 'second' failed validation")
    exc.match(r"'min' failed for 1 rows")



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

from pydiverse.colspec.optional_dependency import assert_frame_equal, pdt


def assert_table_equal(t1: pdt.Table, t2: pdt.Table):
    assert_frame_equal(t1 >> pdt.export(pdt.Polars()), t2 >> pdt.export(pdt.Polars()))



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

import functools
import operator
from dataclasses import dataclass

import pytest

import pydiverse.colspec as cs
import pydiverse.colspec.collection
from pydiverse.colspec.exc import MemberValidationError
from pydiverse.colspec.optional_dependency import C, ColExpr, pdt
from pydiverse.colspec.pdt_util import num_rows
from pydiverse.colspec.testing.assert_equal import assert_table_equal

# ------------------------------------------------------------------------------------ #
#                                        SCHEMA                                        #
# ------------------------------------------------------------------------------------ #


class MyFirstColSpec(cs.ColSpec):
    a = cs.Integer(primary_key=True)
    b = cs.Integer()


class MySecondColSpec(cs.ColSpec):
    a = cs.Integer(primary_key=True)
    b = cs.Integer(min=1)


@dataclass
class MyCollection(pydiverse.colspec.collection.Collection):
    first: MyFirstColSpec
    second: MySecondColSpec

    @cs.filter()
    def equal_primary_keys(self) -> ColExpr:
        return functools.reduce(
            operator.and_,
            (self.first[key] == self.second[key] for key in self.common_primary_keys()),
        )

    @cs.filter()
    def first_b_greater_second_b(self) -> ColExpr:
        return (
            (self.first.b > self.second.b)
            | self.pk_is_null(self.first)
            | self.pk_is_null(self.second)
        )


@dataclass
class SimpleCollection(pydiverse.colspec.collection.Collection):
    first: MyFirstColSpec
    second: MySecondColSpec


def to_my_collection(c: SimpleCollection) -> MyCollection:
    return MyCollection(c.first, c.second)


# ------------------------------------------------------------------------------------ #
#                                         TESTS                                        #
# ------------------------------------------------------------------------------------ #


@pytest.fixture()
def data_without_filter_without_rule_violation() -> SimpleCollection:
    c = SimpleCollection.build()
    c.first = pdt.Table({"a": [1, 2, 3], "b": [1, 2, 3]})
    c.second = pdt.Table({"a": [1, 2, 3], "b": [1, 2, 3]})
    return c


@pytest.fixture()
def data_without_filter_with_rule_violation() -> SimpleCollection:
    c = SimpleCollection.build()
    c.first = pdt.Table({"a": [1, 2, 1], "b": [1, 2, 3]})
    c.second = pdt.Table({"a": [1, 2, 3], "b": [0, 1, 2]})
    return c


@pytest.fixture()
def data_with_filter_without_rule_violation() -> SimpleCollection:
    c = SimpleCollection.build()
    c.first = pdt.Table({"a": [1, 2, 3], "b": [1, 1, 3]})
    c.second = pdt.Table({"a": [2, 3, 4, 5], "b": [1, 2, 3, 4]})
    return c


@pytest.fixture()
def data_with_filter_with_rule_violation() -> SimpleCollection:
    c = SimpleCollection.build()
    c.first = pdt.Table({"a": [1, 2, 3], "b": [1, 2, 3]})
    c.second = pdt.Table({"a": [2, 3, 4, 5], "b": [0, 1, 2, 3]})
    return c


# -------------------------------------- FILTER -------------------------------------- #


@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
def test_filter_without_filter_without_rule_violation(
    data_without_filter_without_rule_violation: SimpleCollection,
):
    out, failure = data_without_filter_without_rule_violation.filter()

    assert isinstance(out, SimpleCollection)
    assert_table_equal(out.first, data_without_filter_without_rule_violation.first)
    assert_table_equal(out.second, data_without_filter_without_rule_violation.second)
    assert len(failure.first) == 0
    assert len(failure.second) == 0


@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
def test_filter_without_filter_with_rule_violation(
    data_without_filter_with_rule_violation: SimpleCollection,
):
    out, failure = data_without_filter_with_rule_violation.filter()

    assert isinstance(out, SimpleCollection)
    assert num_rows(out.first) == 1
    assert num_rows(out.second) == 2
    assert failure.first.counts() == {"_primary_key_": 2}
    assert failure.second.counts() == {"b|min": 1}


@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
def test_filter_with_filter_without_rule_violation(
    data_with_filter_without_rule_violation: SimpleCollection,
):
    my_collection = to_my_collection(data_with_filter_without_rule_violation)
    out, failure = my_collection.filter()

    assert isinstance(out, MyCollection)
    assert_table_equal(out.first, pdt.Table({"a": [3], "b": [3]}))
    assert_table_equal(out.second, pdt.Table({"a": [3], "b": [2]}))
    assert failure.first.counts() == {
        "equal_primary_keys": 1,
        "first_b_greater_second_b": 1,
    }
    assert failure.second.counts() == {
        "equal_primary_keys": 2,
        "first_b_greater_second_b": 1,
    }


@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
def test_filter_with_filter_with_rule_violation(
    data_with_filter_with_rule_violation: SimpleCollection,
):
    my_collection = to_my_collection(data_with_filter_with_rule_violation)
    out, failure = my_collection.filter()

    assert isinstance(out, MyCollection)
    assert_table_equal(out.first, pdt.Table({"a": [3], "b": [3]}))
    assert_table_equal(out.second, pdt.Table({"a": [3], "b": [1]}))
    assert failure.first.counts() == {"equal_primary_keys": 2}
    assert failure.second.counts() == {"b|min": 1, "equal_primary_keys": 2}


# -------------------------------- VALIDATE WITH DATA -------------------------------- #


@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
def test_validate_without_filter_without_rule_violation(
    data_without_filter_without_rule_violation: SimpleCollection,
):
    assert data_without_filter_without_rule_violation.is_valid()
    out = data_without_filter_without_rule_violation.validate()

    assert isinstance(out, SimpleCollection)
    assert_table_equal(out.first, data_without_filter_without_rule_violation.first)
    assert_table_equal(out.second, data_without_filter_without_rule_violation.second)


@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
def test_validate_without_filter_with_rule_violation(
    data_without_filter_with_rule_violation: SimpleCollection,
):
    assert not data_without_filter_with_rule_violation.is_valid()

    with pytest.raises(
        MemberValidationError, match=r"2 members failed validation"
    ) as exc:
        data_without_filter_with_rule_violation.validate()

    exc.match(r"Member 'first' failed validation")
    exc.match(r"'_primary_key_' failed validation for 2 rows")
    exc.match(r"Member 'second' failed validation")
    exc.match(r"Column 'b' failed validation for 1 rules:")
    exc.match(r"'min' failed for 1 rows")


@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
def test_validate_with_filter_without_rule_violation(
    data_with_filter_without_rule_violation: SimpleCollection,
):
    my_collection = to_my_collection(data_with_filter_without_rule_violation)

    assert not my_collection.is_valid()

    with pytest.raises(
        MemberValidationError, match=r"2 members failed validation"
    ) as exc:
        my_collection.validate()

    exc.match(r"Member 'first' failed validation")
    exc.match(r"'equal_primary_keys' failed validation for 1 rows")
    exc.match(r"'first_b_greater_second_b' failed validation for 1 rows")
    exc.match(r"Member 'second' failed validation")
    exc.match(r"'equal_primary_keys' failed validation for 2 rows")


@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
def test_validate_with_filter_with_rule_violation(
    data_with_filter_with_rule_violation: SimpleCollection,
):
    my_collection = to_my_collection(data_with_filter_with_rule_violation)
    assert not my_collection.is_valid()

    with pytest.raises(
        MemberValidationError, match=r"2 members failed validation"
    ) as exc:
        my_collection.validate()

    exc.match(r"Member 'first' failed validation")
    exc.match(r"'equal_primary_keys' failed validation for 2 rows")
    exc.match(r"Member 'second' failed validation")
    exc.match(r"'min' failed for 1 rows")



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

from collections.abc import Callable

import pydiverse.common as pdc

from ..optional_dependency import ColExpr
from ._base import Column
from .struct import Struct


class List(Column):
    """A list column."""

    def __init__(
        self,
        inner: Column,
        *,
        nullable: bool = True,
        primary_key: bool = False,
        check: Callable[[ColExpr], ColExpr] | None = None,
        alias: str | None = None,
        min_length: int | None = None,
        max_length: int | None = None,
    ):
        """
        Args:
            inner: The inner column type. If this type has ``primary_key=True`` set, all
                list items are required to be unique. If the inner type is a struct and
                any of the struct fields have ``primary_key=True`` set, these fields
                must be unique across all list items. Note that if the struct itself has
                ``primary_key=True`` set, the fields' settings do not take effect.
            nullable: Whether this column may contain null values.
            primary_key: Whether this column is part of the primary key of the schema.
            check: A custom check to run for this column. Must return a non-aggregated
                boolean expression.
            alias: An overwrite for this column's name which allows for using a column
                name that is not a valid Python identifier. Especially note that setting
                this option does _not_ allow to refer to the column with two different
                names, the specified alias is the only valid name.
        """
        super().__init__(
            nullable=nullable, primary_key=primary_key, check=check, alias=alias
        )
        self.inner = inner
        self.min_length = min_length
        self.max_length = max_length

    def dtype(self) -> pdc.List:
        return pdc.List(self.inner.dtype())

    def validation_rules(self, expr: ColExpr) -> dict[str, ColExpr]:
        raise NotImplementedError()

        from pydiverse.colspec.optional_dependency import (
            pl,  # needs to be replaced with pydiverse.transform
        )

        inner_rules = {
            f"inner_{rule_name}": expr.list.eval(inner_expr).list.all()
            for rule_name, inner_expr in self.inner.validation_rules(
                pl.element()
            ).items()
        }

        list_rules: dict[str, ColExpr] = {}
        if self.inner.primary_key:
            list_rules["primary_key"] = ~expr.list.eval(
                pl.element().is_duplicated()
            ).list.any()
        elif isinstance(self.inner, Struct) and any(
            col.primary_key for col in self.inner.inner.values()
        ):
            primary_key_columns = [
                name for name, col in self.inner.inner.items() if col.primary_key
            ]
            # NOTE: We optimize for a single primary key column here as it is much
            #  faster to run duplication checks for non-struct types in polars 1.22.
            if len(primary_key_columns) == 1:
                list_rules["primary_key"] = ~expr.list.eval(
                    pl.element().struct.field(primary_key_columns[0]).is_duplicated()
                ).list.any()
            else:
                list_rules["primary_key"] = ~expr.list.eval(
                    pl.struct(
                        pl.element().struct.field(primary_key_columns)
                    ).is_duplicated()
                ).list.any()

        if self.min_length is not None:
            list_rules["min_length"] = (
                pl.when(expr.is_null())
                .then(pl.lit(None))
                .otherwise(expr.list.len() >= self.min_length)
            )
        if self.max_length is not None:
            list_rules["max_length"] = (
                pl.when(expr.is_null())
                .then(pl.lit(None))
                .otherwise(expr.list.len() <= self.max_length)
            )
        return {
            **super().validation_rules(expr),
            **list_rules,
            **inner_rules,
        }



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

import pydiverse.colspec as cs
import pydiverse.colspec.collection
from pydiverse.colspec.optional_dependency import dy, pl

pytestmark = pytest.mark.skipif(
    dy.Column is None, reason="dataframely is required for this test"
)


class TestColSpec(cs.ColSpec):
    a = cs.Integer()


class MyCollection(pydiverse.colspec.collection.Collection):
    first: TestColSpec
    second: TestColSpec | None


def test_collection_missing_required_member():
    with pytest.raises(ValueError):
        MyCollection.validate_polars_data({"second": pl.LazyFrame({"a": [1, 2, 3]})})


def test_collection_superfluous_member():
    with pytest.warns(Warning):
        MyCollection.validate_polars_data(
            {
                "first": pl.LazyFrame({"a": [1, 2, 3]}),
                "third": pl.LazyFrame({"a": [1, 2, 3]}),
            },
        )



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import inspect
import types
import typing
from collections.abc import Iterable, Mapping
from functools import reduce
from typing import Any, Literal, Self, overload

from pydiverse.colspec._validation import validate_columns, validate_dtypes
from pydiverse.colspec.columns._base import Column

from . import GroupRule, GroupRulePolars, Rule, RulePolars, exc
from .config import Config, alias_subquery
from .exc import (
    ImplementationError,
    RuleValidationError,
    ValidationError,
)
from .failure import FailureInfo
from .optional_dependency import ColExpr, Generator, dag, dy, pa, pdt, pl, sa

_ORIGINAL_NULL_SUFFIX = "__orig_null__"


def convert_to_dy_col_spec(base_class):
    from pydiverse.colspec import Column

    if base_class == ColSpec:
        # stop base class iteration
        return {}
    elif inspect.isclass(base_class) and issubclass(base_class, ColSpec):
        dy_cols = (
            {
                k: convert_to_dy(v)
                for k, v in base_class.__dict__.items()
                if isinstance(v, Column)
            }
            | {
                k: convert_to_dy(v())
                for k, v in base_class.__dict__.items()
                if inspect.isclass(v) and issubclass(v, Column)
            }
            | {
                k: dy._rule.GroupRule(v.expr, v.group_columns)
                if isinstance(v, GroupRulePolars)
                else dy._rule.Rule(v.expr)
                for k, v in base_class.__dict__.items()
                if isinstance(v, RulePolars)
            }
        )
        for base in base_class.__bases__:
            dy_cols.update(convert_to_dy_col_spec(base))
        return dy_cols
    else:
        return {}


def convert_to_dy_anno(annotation):
    if isinstance(annotation, types.UnionType):
        anno_types = [convert_to_dy_anno(t) for t in typing.get_args(annotation)]
        return reduce(lambda x, y: x | y, anno_types)
    if inspect.isclass(annotation) and issubclass(annotation, ColSpec):
        col_spec = convert_to_dy_col_spec(annotation)
        return dy.LazyFrame[type(annotation.__name__, (dy.Schema,), col_spec.copy())]
    else:
        return annotation


def convert_to_dy_anno_dict(annotations: dict[str, typing.Any]):
    return {k: convert_to_dy_anno(v) for k, v in annotations.items()}


def convert_to_dy(value):
    from pydiverse.colspec import Column

    if isinstance(value, Column) and hasattr(dy, value.__class__.__name__):
        return value.to_dataframely()
    else:
        return value


class ColSpecMeta(type):
    def __new__(cls, clsname, bases, attribs):
        # change bases (only ABC is a real base)
        bases = bases[0:1]
        return super().__new__(cls, clsname, bases, attribs)


class ColSpecBase:
    #  this is just the same as object but ruff does not kill it
    pass


class ColSpec(
    ColSpecBase,
    FailureInfo,
    pdt.Table,
    dag.Table,
    pl.LazyFrame,
    pl.DataFrame,
    metaclass=ColSpecMeta,
):
    """Base class for all column specifications.

    The base classes here are just for code completion support when working with
    Collection objects that store actual data or table references. They are removed
    at runtime by a metaclass.
    """

    def __init__(self):
        pass

    @classmethod
    def get_subquery_name(cls, tbl: pdt.Table, rule_name: str) -> str:
        """Get the name of the subquery for a given rule.

        Args:
            tbl: The table to which the rule applies.
            rule_name: The name of the rule.

        Returns:
            The name of the subquery.
        """
        if tbl._ast.name is not None and tbl._ast.name != "<unnamed>":
            return f"{tbl._ast.name}_{rule_name}"
        else:
            return rule_name

    @classmethod
    def fail_dy_columns_in_colspec(cls):
        if dy.Column is not None:
            if any(
                [
                    isinstance(getattr(cls, c), dy.Column | dy._rule.Rule)
                    for c in dir(cls)
                ]
            ):
                if any([isinstance(getattr(cls, c), dy.Column) for c in dir(cls)]):
                    raise ImplementationError(
                        "Dataframely Columns won't work in ColSpec classes. Most likely"
                        " you find the same column classes in pydiverse.colspec. With "
                        "import pydiverse.colspec as cs, `dy.Integer` becomes "
                        "`cs.Integer` for example."
                    )
                else:
                    raise ImplementationError(
                        "Dataframely Rules won't work in ColSpec classes. You can use "
                        "`@cs.rule` decorator instead of `@dy.rule` decorator in case "
                        "you import pydiverse.colspec as cs."
                    )
        if any(
            [
                isinstance(getattr(cls, c), staticmethod)
                and isinstance(getattr(cls, c).__wrapped__, Rule | RulePolars)
                for c in dir(cls)
            ]
        ):
            # TODO: better message showing member which causes error
            raise ImplementationError(
                "The @staticmethod decorator needs to be after @cs.rule decorator."
            )

    @classmethod
    def primary_keys(cls) -> list[str]:
        """Returns a list of column names that are marked as primary keys.

        Returns:
            list[str]: Names of columns that are primary keys
        """
        from pydiverse.colspec import Column

        cls.fail_dy_columns_in_colspec()
        result = [
            member
            for member in dir(cls)
            if isinstance(getattr(cls, member), Column)
            and getattr(cls, member).primary_key
        ]
        return result

    @classmethod
    def column_names(cls) -> list[str]:
        cls.fail_dy_columns_in_colspec()
        result = [
            getattr(cls, member).alias or member
            for member in dir(cls)
            if isinstance(getattr(cls, member), Column)
        ]
        return result

    @classmethod
    def columns(cls) -> dict[str, Column]:
        cls.fail_dy_columns_in_colspec()
        return {
            col.alias or name: col
            for name, col in cls.__dict__.items()
            if isinstance(col, Column)
        }

    @classmethod
    def alias_map(cls) -> dict[str, str]:
        return {
            col.alias or name: name
            for name, col in cls.__dict__.items()
            if isinstance(col, Column)
        }

    @classmethod
    def validate(cls, tbl: pdt.Table, cast: bool = False) -> pdt.Table:
        valid_rows, failure = cls.filter(tbl, cast=cast)
        if len(failure) > 0:
            raise RuleValidationError(failure.counts())
        return valid_rows

    @classmethod
    def validate_polars(
        cls, data: pl.DataFrame | pl.LazyFrame, cast: bool = False
    ) -> pl.DataFrame | pl.LazyFrame:
        import dataframely.exc as dy_exc

        dy_schema_cols = convert_to_dy_col_spec(cls)
        try:
            dy_schema = type[dy.Schema](
                cls.__name__, (dy.Schema,), dy_schema_cols.copy()
            )
            return dy_schema.validate(data, cast=cast)
        except (dy_exc.ValidationError, dy_exc.ImplementationError) as e:
            err_type = getattr(exc, e.__class__.__name__)
            f = err_type.__new__(err_type)
            for c in dir(e):
                if not c.startswith("_"):
                    setattr(f, c, getattr(e, c))
            # f.__dict__.update(e.__dict__)
            raise f from e

    @classmethod
    def is_valid(cls, tbl: pdt.Table, *, cast: bool = False) -> bool:
        """Utility method to check whether :meth:`validate` raises an exception.

        Args:
            tbl: The table to check for validity.
            cast: Whether columns with a wrong data type in the input data frame are
                cast to the schema's defined data type before running validation. If set
                to ``False``, a wrong data type will result in a return value of
                ``False``.

        Returns:
            Whether the provided dataframe can be validated with this schema.
        """
        try:
            from polars.exceptions import (
                InvalidOperationError as PlInvalidOperationError,
            )
        except ImportError:
            PlInvalidOperationError = None

        try:
            cls.validate(tbl, cast=cast)
            return True
        except (ValidationError, PlInvalidOperationError):
            return False
        except Exception as e:  # pragma: no cover
            raise e

    @classmethod
    def is_valid_polars(
        cls, df: pl.DataFrame | pl.LazyFrame, *, cast: bool = False
    ) -> bool:
        """Utility method to check whether :meth:`validate` raises an exception.

        Args:
            df: The data frame to check for validity.
            cast: Whether columns with a wrong data type in the input data frame are
                cast to the schema's defined data type before running validation. If set
                to ``False``, a wrong data type will result in a return value of
                ``False``.

        Returns:
            Whether the provided dataframe can be validated with this schema.
        """
        import polars.exceptions as plexc

        try:
            cls.validate_polars(df, cast=cast)
            return True
        except (ValidationError, plexc.InvalidOperationError):
            return False
        except Exception as e:  # pragma: no cover
            raise e

    @classmethod
    def sample_polars(
        cls,
        num_rows: int = 1,
        generator: Generator | None = None,
        *,
        overrides: Mapping[str, Iterable[Any]] | None = None,
    ) -> pl.DataFrame | pl.LazyFrame:
        dy_schema_cols = convert_to_dy_col_spec(cls)
        dy_schema: type[dy.Schema] = type[dy.Schema](
            cls.__name__, (dy.Schema,), dy_schema_cols.copy()
        )
        return dy_schema.sample(num_rows, generator=generator, overrides=overrides)

    @classmethod
    def create_empty_polars(cls) -> dy.DataFrame[Self]:
        dy_schema_cols = convert_to_dy_col_spec(cls)
        dy_schema = type[dy.Schema](cls.__name__, (dy.Schema,), dy_schema_cols.copy())
        return dy_schema.create_empty()

    # ------------------------------------ CASTING ----------------------------------- #

    @overload
    @classmethod
    def cast_polars(
        cls, df: pl.DataFrame
    ) -> dy.DataFrame[Self]: ...  # pragma: no cover

    @overload
    @classmethod
    def cast_polars(
        cls, df: pl.LazyFrame
    ) -> dy.LazyFrame[Self]: ...  # pragma: no cover

    @classmethod
    def cast_polars(
        cls, df: pl.DataFrame | pl.LazyFrame
    ) -> dy.DataFrame[Self] | dy.LazyFrame[Self]:
        dy_schema_cols = convert_to_dy_col_spec(cls)
        dy_schema = type[dy.Schema](cls.__name__, (dy.Schema,), dy_schema_cols.copy())
        return dy_schema.cast(df)

    @classmethod
    def polars_schema(cls) -> pl.Schema:
        return pl.Schema(
            {
                name: getattr(cls, name).dtype().to_polars()
                for name in dir(cls)
                if isinstance(getattr(cls, name), Column)
            }
        )

    # ----------------------------------- FILTERING ---------------------------------- #

    @classmethod
    def filter(
        cls,
        tbl: pdt.Table,
        *,
        cast: bool = False,
        cfg: Config = Config.default,
    ) -> tuple[Self, FailureInfo]:
        """Filter the table by the rules of this column specification.

        This method can be thought of as a "soft alternative" to :meth:`validate`.
        While :meth:`validate` raises an exception when a row does not adhere to the
        rules defined in the schema, this method simply filters out these rows and
        succeeds.

        Args:
            tbl: The data frame to filter for valid rows. The data frame is collected
                within this method, regardless of whether a :class:`~polars.DataFrame`
                or :class:`~polars.LazyFrame` is passed.
            cast: Whether columns with a wrong data type in the input data frame are
                cast to the schema's defined data type if possible. Rows for which the
                cast fails for any column are filtered out.

        Returns:
            A tuple of the validated rows in the input data frame (potentially
            empty) and a simple dataclass carrying information about the rows of the
            data frame which could not be validated successfully.

        Raises:
            ValidationError: If the columns of the input data frame are invalid. This
                happens only if the data frame misses a column defined in the schema or
                a column has an invalid dtype while ``cast`` is set to ``False``.

        Note:
            This method preserves the ordering of the input data frame.
        """

        src_tbl = tbl

        tbl = cls._validate_columns(tbl, casting=("lenient" if cast else "none"))
        rules, group_rules = cls._validation_rules(tbl)

        if cast:
            dtype_rules = {
                f"{col}|dtype": (
                    tbl[col].is_null() == tbl[f"{col}{_ORIGINAL_NULL_SUFFIX}"]
                )
                for col in cls.column_names()
            }
            rules.update(dtype_rules)

        if "_primary_key_" in rules or "_primary_key_" in group_rules:
            raise ImplementationError(
                "@cs.rule annotated functions must not be called `_primary_key_`"
            )
        if len(cls.primary_keys()) > 0:
            group_rules["_primary_key_"] = GroupRule(
                pdt.count() == 1, group_columns=cls.primary_keys()
            )

        for name, group_rule in group_rules.items():
            subquery = (
                tbl
                >> pdt.group_by(*group_rule.group_columns)
                >> pdt.summarize(expr=group_rule.expr)
                >> alias_subquery(cfg, cls.get_subquery_name(tbl, name))
            )
            tbl >>= pdt.left_join(
                subquery,
                on=pdt.all(
                    *[tbl[col] == subquery[col] for col in group_rule.group_columns]
                ),
            ) >> pdt.select(*tbl)
            rules[name] = subquery.expr

        combined = pdt.all(True, *rules.values())

        if cast:
            # Rules other than the "dtype rule" might not be reliable if type casting
            # failed, i.e. if the "dtype rule" evaluated to `False`. For this reason,
            # we set all other rule evaluations to `null` in the case of dtype casting
            # failure.
            # TODO: actually we should only do this for expressions containing a column
            # that failed the cast.
            all_dtype_casts_valid = pdt.all(
                True,
                *(
                    tbl[col].is_null() == tbl[f"{col}{_ORIGINAL_NULL_SUFFIX}"]
                    for col in cls.column_names()
                ),
            )

            # remove original null information again
            tbl >>= pdt.drop(
                *(tbl[f"{col}{_ORIGINAL_NULL_SUFFIX}"] for col in cls.column_names())
            )

            rules.update(
                {
                    name: pdt.when(all_dtype_casts_valid)
                    .then(expr)
                    .otherwise(pdt.lit(None, dtype=pdt.Bool))
                    for name, expr in rules.items()
                    if not name.endswith("|dtype")
                }
            )

        ok_rows = tbl >> pdt.filter(combined)
        invalid_rows = tbl >> pdt.filter(~combined) >> pdt.mutate(**rules)

        if len(cls.primary_keys()) > 0:
            ok_rows = ok_rows
            invalid_rows = invalid_rows

        return ok_rows, FailureInfo(
            tbl=src_tbl,
            invalid_rows=invalid_rows,
            rule_columns=rules,
            cfg=cfg,
        )

    @classmethod
    def _validate_columns(
        cls, tbl: pdt.Table, *, casting: Literal["none", "lenient", "strict"]
    ):
        cls.fail_dy_columns_in_colspec()

        tbl = validate_columns(tbl, expected=cls.column_names())

        if casting == "lenient":
            tbl >>= pdt.mutate(
                **{f"{col.name}{_ORIGINAL_NULL_SUFFIX}": col.is_null() for col in tbl}
            )

        return validate_dtypes(tbl, expected=cls.columns(), casting=casting)

    @classmethod
    def _validation_rules(
        cls, tbl: pdt.Table
    ) -> tuple[dict[str, ColExpr], dict[str, GroupRule]]:
        cls.fail_dy_columns_in_colspec()
        alias_map = cls.alias_map()
        return {
            f"{col}|{rule_name}": rule.fill_null(True)
            for col in cls.column_names()
            for rule_name, rule in getattr(cls, alias_map[col])
            .validation_rules(tbl[col])
            .items()
        } | {
            rule: getattr(cls, rule).expr
            for rule in dir(cls)
            if isinstance(getattr(cls, rule), Rule)
        }, {
            rule: getattr(cls, rule)
            for rule in dir(cls)
            if isinstance(getattr(cls, rule), GroupRule)
        }

    @classmethod
    def filter_polars(
        cls, df: pl.DataFrame | pl.LazyFrame, *, cast: bool = False
    ) -> tuple[dy.DataFrame[Self], dy.FailureInfo]:
        """Filter the data frame by the rules of this schema.

        This method can be thought of as a "soft alternative" to :meth:`validate`.
        While :meth:`validate` raises an exception when a row does not adhere to the
        rules defined in the schema, this method simply filters out these rows and
        succeeds.

        Args:
            df: The data frame to filter for valid rows. The data frame is collected
                within this method, regardless of whether a :class:`~polars.DataFrame`
                or :class:`~polars.LazyFrame` is passed.
            cast: Whether columns with a wrong data type in the input data frame are
                cast to the schema's defined data type if possible. Rows for which the
                cast fails for any column are filtered out.

        Returns:
            A tuple of the validated rows in the input data frame (potentially
            empty) and a simple dataclass carrying information about the rows of the
            data frame which could not be validated successfully.

        Raises:
            ValidationError: If the columns of the input data frame are invalid. This
                happens only if the data frame misses a column defined in the schema or
                a column has an invalid dtype while ``cast`` is set to ``False``.

        Note:
            This method preserves the ordering of the input data frame.
        """
        import dataframely.exc as dy_exc

        dy_schema_cols = convert_to_dy_col_spec(cls)
        dy_schema = type[dy.Schema](cls.__name__, (dy.Schema,), dy_schema_cols.copy())

        try:
            return dy_schema.filter(df, cast=cast)
        except (dy_exc.ValidationError, dy_exc.ImplementationError) as e:
            err_type = getattr(exc, e.__class__.__name__)
            f = err_type.__new__(err_type)
            f.__dict__.update(e.__dict__)
            raise f from e

    @classmethod
    def sql_schema(cls, dialect: sa.Dialect) -> list[sa.Column]:
        """Obtain the SQL schema for a particular dialect for this schema.

        Args:
            dialect: The dialect for which to obtain the SQL schema. Note that column
                datatypes may differ across dialects.

        Returns:
            A list of :mod:`sqlalchemy` columns that can be used to create a table
            with the schema as defined by this class.
        """
        return [
            col.sqlalchemy_column(name, dialect) for name, col in cls.columns().items()
        ]

    @classmethod
    def pyarrow_schema(cls) -> pa.Schema:
        """Obtain the pyarrow schema for this schema.

        Returns:
            A :mod:`pyarrow` schema that mirrors the schema defined by this class.
        """
        return pa.schema(
            [col.pyarrow_field(name) for name, col in cls.columns().items()]
        )



# Copyright (c) QuantCo and pydiverse contributors 2025-2025
# SPDX-License-Identifier: BSD-3-Clause

import datetime as dt
import decimal
import math
from typing import Any, cast

import pytest

import pydiverse.colspec as cs
from pydiverse.colspec import Column
from pydiverse.colspec.optional_dependency import Generator, dy, pl
from pydiverse.colspec.testing import (
    COLUMN_TYPES,
    INTEGER_COLUMN_TYPES,
    SUPERTYPE_COLUMN_TYPES,
    create_colspec,
)


@pytest.fixture()
def generator() -> Generator:
    return Generator(seed=42)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize("column_type", COLUMN_TYPES + SUPERTYPE_COLUMN_TYPES)
def test_sample_custom_check(column_type: type[Column], generator: Generator):
    column = column_type(check=lambda expr: expr)
    with pytest.raises(ValueError):
        column.sample_polars(generator)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize("column_type", COLUMN_TYPES + SUPERTYPE_COLUMN_TYPES)
@pytest.mark.parametrize("nullable", [True, False])
def test_sample_valid(column_type: type[Column], nullable: bool, generator: Generator):
    column = column_type(nullable=nullable)
    samples = sample_and_validate(column, generator, n=10_000)
    if nullable:
        assert math.isclose(cast(float, samples.is_null().mean()), 0.1, abs_tol=0.01)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_sample_any(generator: Generator):
    column = cs.Any()
    samples = sample_and_validate(column, generator, n=100)
    assert samples.is_null().all()


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize("column_type", INTEGER_COLUMN_TYPES)
@pytest.mark.parametrize("min_kwargs", [{}, {"min": 10}, {"min_exclusive": 10}])
@pytest.mark.parametrize("max_kwargs", [{}, {"max": 100}, {"max_exclusive": 100}])
def test_sample_integer_min_max(
    column_type: type[Column],
    min_kwargs: dict[str, Any],
    max_kwargs: dict[str, Any],
    generator: Generator,
):
    column = column_type(**min_kwargs, **max_kwargs)
    samples = sample_and_validate(column, generator, n=10_000)
    if min_kwargs and max_kwargs:
        assert samples.min() == (
            min_kwargs["min"]
            if "min" in min_kwargs
            else min_kwargs["min_exclusive"] + 1
        )
        assert samples.max() == (
            max_kwargs["max"]
            if "max" in max_kwargs
            else max_kwargs["max_exclusive"] - 1
        )


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize("column_type", INTEGER_COLUMN_TYPES)
def test_sample_integer_is_in(column_type: type[Column], generator: Generator):
    column = column_type(is_in=[4, 5, 6])  # type: ignore
    samples = sample_and_validate(column, generator, n=10_000)
    assert math.isclose(samples.mean(), 5, abs_tol=0.1)  # type: ignore
    assert samples.min() == 4
    assert samples.max() == 6


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize(
    "column",
    [
        cs.String(regex=".*", min_length=1),
        cs.String(regex=".*", max_length=2),
        cs.String(regex=".*", min_length=1, max_length=5),
    ],
)
def test_sample_string_invalid(column: Column, generator: Generator):
    with pytest.raises(ValueError):
        column.sample_polars(generator)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize(
    "column",
    [
        cs.String(),
        cs.String(min_length=3),
        cs.String(max_length=5),
        cs.String(min_length=3, max_length=5),
        cs.String(regex="[abc]def(ghi)?"),
    ],
)
def test_sample_string(column: Column, generator: Generator):
    sample_and_validate(column, generator, n=10_000)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_sample_decimal(generator: Generator):
    column = cs.Decimal(precision=3, scale=2, max_exclusive=decimal.Decimal("6.5"))
    samples = sample_and_validate(column, generator, n=100_000)
    assert samples.min() == decimal.Decimal("-9.99")
    assert samples.max() == decimal.Decimal("6.49")


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_sample_date(generator: Generator):
    column = cs.Date(
        min=dt.date(2020, 1, 1), max=dt.date(2021, 12, 1), resolution="1mo"
    )
    samples = sample_and_validate(column, generator, n=100_000)
    assert samples.min() == dt.date(2020, 1, 1)
    assert samples.max() == dt.date(2021, 12, 1)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_sample_date_9999(generator: Generator):
    column = cs.Date(
        min=dt.date(9998, 1, 1), max=dt.date(9999, 12, 1), resolution="1mo"
    )
    samples = sample_and_validate(column, generator, n=100_000)
    assert samples.min() == dt.date(9998, 1, 1)
    assert samples.max() == dt.date(9999, 12, 1)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_sample_datetime(generator: Generator):
    column = cs.Datetime(
        min=dt.datetime(2020, 1, 1),
        max_exclusive=dt.datetime(2022, 1, 1),
        resolution="1d",
    )
    samples = sample_and_validate(column, generator, n=100_000)
    assert samples.min() == dt.datetime(2020, 1, 1)
    assert samples.max() == dt.datetime(2021, 12, 31)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_sample_time(generator: Generator):
    column = cs.Time(min=dt.time(), max=dt.time(23, 59), resolution="1m")
    samples = sample_and_validate(column, generator, n=1_000_000)
    assert samples.min() == dt.time()
    assert samples.max() == dt.time(23, 59)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_sample_duration(generator: Generator):
    column = cs.Duration(
        min=dt.timedelta(hours=24), max=dt.timedelta(hours=120), resolution="12h"
    )
    samples = sample_and_validate(column, generator, n=100_000)
    assert samples.min() == dt.timedelta(hours=24)
    assert samples.max() == dt.timedelta(hours=120)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_sample_enum(generator: Generator):
    column = cs.Enum(["a", "b", "c"], nullable=False)
    samples = sample_and_validate(column, generator, n=10_000)
    assert set(samples) == {"a", "b", "c"}


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_sample_list(generator: Generator):
    column = cs.List(cs.String(regex="[abc]"), min_length=5, max_length=10)
    samples = sample_and_validate(column, generator, n=10_000)
    assert set(samples.list.len()) == set(range(5, 11)) | {None}


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_sample_struct(generator: Generator):
    column = cs.Struct({"a": cs.String(regex="[abc]"), "b": cs.String(regex="[a-z]xx")})
    samples = sample_and_validate(column, generator, n=10_000)
    assert len(samples) == 10_000


# --------------------------------------- UTILS -------------------------------------- #


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def sample_and_validate(column: Column, generator: Generator, *, n: int) -> pl.Series:
    samples = column.sample_polars(generator, n=n)
    col_spec = create_colspec("test", {"a": column})
    col_spec.validate_polars(pl.DataFrame({"a": samples}))
    return samples



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

import datetime as dt
from typing import Any

import pytest

import pydiverse.colspec as cs
from pydiverse.colspec import Column
from pydiverse.colspec.exc import DtypeValidationError
from pydiverse.colspec.optional_dependency import C, Generator, dy, pdt, pl
from pydiverse.colspec.testing.factory import create_colspec
from pydiverse.colspec.testing.rules import evaluate_rules


@pytest.mark.parametrize(
    ("column_type", "kwargs"),
    [
        (cs.Date, {"min": dt.date(2020, 1, 15), "max": dt.date(2020, 1, 14)}),
        (cs.Date, {"min_exclusive": dt.date(2020, 1, 15), "max": dt.date(2020, 1, 15)}),
        (cs.Date, {"min": dt.date(2020, 1, 15), "max_exclusive": dt.date(2020, 1, 15)}),
        (
            cs.Date,
            {
                "min_exclusive": dt.date(2020, 1, 15),
                "max_exclusive": dt.date(2020, 1, 15),
            },
        ),
        (
            cs.Date,
            {"min": dt.date(2020, 1, 15), "min_exclusive": dt.date(2020, 1, 15)},
        ),
        (
            cs.Date,
            {"max": dt.date(2020, 1, 15), "max_exclusive": dt.date(2020, 1, 15)},
        ),
        (
            cs.Datetime,
            {"min": dt.datetime(2020, 1, 15), "max": dt.datetime(2020, 1, 14)},
        ),
        (
            cs.Datetime,
            {
                "min_exclusive": dt.datetime(2020, 1, 15),
                "max": dt.datetime(2020, 1, 15),
            },
        ),
        (
            cs.Datetime,
            {
                "min": dt.datetime(2020, 1, 15),
                "max_exclusive": dt.datetime(2020, 1, 15),
            },
        ),
        (
            cs.Datetime,
            {
                "min_exclusive": dt.datetime(2020, 1, 15),
                "max_exclusive": dt.datetime(2020, 1, 15),
            },
        ),
        (
            cs.Datetime,
            {
                "min": dt.datetime(2020, 1, 15),
                "min_exclusive": dt.datetime(2020, 1, 15),
            },
        ),
        (
            cs.Datetime,
            {
                "max": dt.datetime(2020, 1, 15),
                "max_exclusive": dt.datetime(2020, 1, 15),
            },
        ),
        (cs.Time, {"min": dt.time(12, 15), "max": dt.time(12, 14)}),
        (cs.Time, {"min_exclusive": dt.time(12, 15), "max": dt.time(12, 15)}),
        (cs.Time, {"min": dt.time(12, 15), "max_exclusive": dt.time(12, 15)}),
        (
            cs.Time,
            {
                "min_exclusive": dt.time(12, 15),
                "max_exclusive": dt.time(12, 15),
            },
        ),
        (
            cs.Time,
            {"min": dt.time(12, 15), "min_exclusive": dt.time(12, 15)},
        ),
        (
            cs.Time,
            {"max": dt.time(12, 15), "max_exclusive": dt.time(12, 15)},
        ),
        (
            cs.Duration,
            {"min": dt.timedelta(seconds=15), "max": dt.timedelta(seconds=14)},
        ),
        (
            cs.Duration,
            {
                "min_exclusive": dt.timedelta(seconds=15),
                "max": dt.timedelta(seconds=15),
            },
        ),
        (
            cs.Duration,
            {
                "min": dt.timedelta(seconds=15),
                "max_exclusive": dt.timedelta(seconds=15),
            },
        ),
        (
            cs.Duration,
            {
                "min_exclusive": dt.timedelta(seconds=15),
                "max_exclusive": dt.timedelta(seconds=15),
            },
        ),
        (
            cs.Duration,
            {
                "min": dt.timedelta(seconds=15),
                "min_exclusive": dt.timedelta(seconds=15),
            },
        ),
        (
            cs.Duration,
            {
                "max": dt.timedelta(seconds=15),
                "max_exclusive": dt.timedelta(seconds=15),
            },
        ),
    ],
)
def test_args_consistency_min_max(column_type: type[Column], kwargs: dict[str, Any]):
    with pytest.raises(ValueError):
        column_type(**kwargs)


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize(
    ("column_type", "kwargs"),
    [
        (cs.Date, {"min": dt.date(2020, 1, 10), "resolution": "1mo"}),
        (cs.Date, {"min_exclusive": dt.date(2020, 1, 10), "resolution": "1mo"}),
        (cs.Date, {"max": dt.date(2020, 1, 10), "resolution": "1mo"}),
        (cs.Date, {"max_exclusive": dt.date(2020, 1, 10), "resolution": "1mo"}),
        (cs.Date, {"resolution": "1h"}),
        (cs.Date, {"resolution": "1d6h"}),
        (cs.Datetime, {"min": dt.datetime(2020, 1, 15, 11), "resolution": "1d"}),
        (
            cs.Datetime,
            {"min_exclusive": dt.datetime(2020, 1, 15, 11), "resolution": "1d"},
        ),
        (cs.Datetime, {"max": dt.datetime(2020, 1, 15, 11), "resolution": "1d"}),
        (
            cs.Datetime,
            {"max_exclusive": dt.datetime(2020, 1, 15, 11), "resolution": "1d"},
        ),
        (cs.Time, {"min": dt.time(12, 15), "resolution": "1h"}),
        (cs.Time, {"min_exclusive": dt.time(12, 15), "resolution": "1h"}),
        (cs.Time, {"max": dt.time(12, 15), "resolution": "1h"}),
        (cs.Time, {"max_exclusive": dt.time(12, 15), "resolution": "1h"}),
        (cs.Time, {"resolution": "1d"}),
        (cs.Time, {"resolution": "1d6h"}),
        (cs.Duration, {"min": dt.timedelta(minutes=30), "resolution": "1h"}),
        (cs.Duration, {"min_exclusive": dt.timedelta(minutes=30), "resolution": "1h"}),
        (cs.Duration, {"max": dt.timedelta(minutes=30), "resolution": "1h"}),
        (cs.Duration, {"max_exclusive": dt.timedelta(minutes=30), "resolution": "1h"}),
    ],
)
def test_args_resolution_invalid(column_type: type[Column], kwargs: dict[str, Any]):
    # Resolution errors are not caught by ColSpec. They only are raised when
    # converting to dataframely Schema in polars functions.
    # In Practice, it is recommended to use microsecond resolution for Datetimes,
    # Time, and Durations and second resolution for Dates. SQL Databases can represent
    # these precisions as well.
    class TestColSpec(cs.ColSpec):
        a = column_type(**kwargs)

    with pytest.raises(ValueError):
        # the ValueError is raised when converting ColSpec to dy.Schema
        TestColSpec.validate_polars(pl.DataFrame(dict(a=[])))


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.parametrize(
    ("column_type", "kwargs"),
    [
        (cs.Date, {"min": dt.date(2020, 1, 1), "resolution": "1mo"}),
        (cs.Date, {"min_exclusive": dt.date(2020, 1, 1), "resolution": "1mo"}),
        (cs.Date, {"max": dt.date(2020, 1, 1), "resolution": "1mo"}),
        (cs.Date, {"max_exclusive": dt.date(2020, 1, 1), "resolution": "1mo"}),
        (cs.Date, {"resolution": "1d"}),
        (cs.Date, {"resolution": "1y"}),
        (cs.Datetime, {"min": dt.datetime(2020, 1, 15), "resolution": "1d"}),
        (cs.Datetime, {"min_exclusive": dt.datetime(2020, 1, 15), "resolution": "1d"}),
        (cs.Datetime, {"max": dt.datetime(2020, 1, 15), "resolution": "1d"}),
        (cs.Datetime, {"max_exclusive": dt.datetime(2020, 1, 15), "resolution": "1d"}),
        (cs.Time, {"min": dt.time(12), "resolution": "1h"}),
        (cs.Time, {"min_exclusive": dt.time(12), "resolution": "1h"}),
        (cs.Time, {"max": dt.time(12), "resolution": "1h"}),
        (cs.Time, {"max_exclusive": dt.time(12), "resolution": "1h"}),
        (cs.Time, {"resolution": "6h"}),
        (cs.Time, {"resolution": "15m"}),
        (cs.Duration, {"min": dt.timedelta(hours=3), "resolution": "1h"}),
        (cs.Duration, {"min_exclusive": dt.timedelta(hours=3), "resolution": "1h"}),
        (cs.Duration, {"max": dt.timedelta(hours=3), "resolution": "1h"}),
        (cs.Duration, {"max_exclusive": dt.timedelta(hours=3), "resolution": "1h"}),
    ],
)
def test_args_resolution_valid(column_type: type[Column], kwargs: dict[str, Any]):
    class TestColSpec(cs.ColSpec):
        a = column_type(**kwargs)

    with pytest.raises(DtypeValidationError):
        # this still tests that there is no ValueError when instantiating
        # dy.Column types
        TestColSpec.validate_polars(pl.DataFrame(dict(a=[])))


@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
@pytest.mark.parametrize(
    ("column", "values", "valid"),
    [
        (
            cs.Date(min=dt.date(2020, 4, 1)),
            [dt.date(2020, 3, 31), dt.date(2020, 4, 1), dt.date(9999, 12, 31)],
            {"min": [False, True, True]},
        ),
        (
            cs.Date(min_exclusive=dt.date(2020, 4, 1)),
            [dt.date(2020, 3, 31), dt.date(2020, 4, 1), dt.date(9999, 12, 31)],
            {"min_exclusive": [False, False, True]},
        ),
        (
            cs.Date(max=dt.date(2020, 4, 1)),
            [dt.date(2020, 3, 31), dt.date(2020, 4, 1), dt.date(2020, 4, 2)],
            {"max": [True, True, False]},
        ),
        (
            cs.Date(max_exclusive=dt.date(2020, 4, 1)),
            [dt.date(2020, 3, 31), dt.date(2020, 4, 1), dt.date(2020, 4, 2)],
            {"max_exclusive": [True, False, False]},
        ),
        (
            cs.Time(min=dt.time(3)),
            [dt.time(2, 59), dt.time(3, 0, 0), dt.time(4)],
            {"min": [False, True, True]},
        ),
        (
            cs.Time(min_exclusive=dt.time(3)),
            [dt.time(2, 59), dt.time(3, 0, 0), dt.time(4)],
            {"min_exclusive": [False, False, True]},
        ),
        (
            cs.Time(max=dt.time(11, 59, 59, 999999)),
            [dt.time(11), dt.time(12), dt.time(13)],
            {"max": [True, False, False]},
        ),
        (
            cs.Time(max_exclusive=dt.time(12)),
            [dt.time(11), dt.time(12), dt.time(13)],
            {"max_exclusive": [True, False, False]},
        ),
        (
            cs.Datetime(min=dt.datetime(2020, 3, 1, hour=12)),
            [
                dt.datetime(2020, 2, 29, hour=14),
                dt.datetime(2020, 3, 1, hour=11),
                dt.datetime(2020, 3, 1, hour=12),
                dt.datetime(2020, 3, 1, hour=18),
                dt.datetime(2020, 3, 2, hour=11),
            ],
            {"min": [False, False, True, True, True]},
        ),
        (
            cs.Datetime(min_exclusive=dt.datetime(2020, 3, 1, hour=12)),
            [
                dt.datetime(2020, 2, 29, hour=14),
                dt.datetime(2020, 3, 1, hour=11),
                dt.datetime(2020, 3, 1, hour=12),
                dt.datetime(2020, 3, 1, hour=18),
                dt.datetime(2020, 3, 2, hour=11),
            ],
            {"min_exclusive": [False, False, False, True, True]},
        ),
        (
            cs.Datetime(max=dt.datetime(2020, 3, 1, hour=12)),
            [
                dt.datetime(2020, 2, 29, hour=14),
                dt.datetime(2020, 3, 1, hour=11),
                dt.datetime(2020, 3, 1, hour=12),
                dt.datetime(2020, 3, 1, hour=18),
                dt.datetime(2020, 3, 2, hour=11),
            ],
            {"max": [True, True, True, False, False]},
        ),
        (
            cs.Datetime(max_exclusive=dt.datetime(2020, 3, 1, hour=12)),
            [
                dt.datetime(2020, 2, 29, hour=14),
                dt.datetime(2020, 3, 1, hour=11),
                dt.datetime(2020, 3, 1, hour=12),
                dt.datetime(2020, 3, 1, hour=18),
                dt.datetime(2020, 3, 2, hour=11),
            ],
            {"max_exclusive": [True, True, False, False, False]},
        ),
        (
            cs.Duration(min=dt.timedelta(days=1, seconds=14400)),
            [
                dt.timedelta(seconds=13000),
                dt.timedelta(days=1, seconds=14400),
                dt.timedelta(days=2),
            ],
            {"min": [False, True, True]},
        ),
        (
            cs.Duration(min_exclusive=dt.timedelta(days=1, seconds=14400)),
            [
                dt.timedelta(seconds=13000),
                dt.timedelta(days=1, seconds=14400),
                dt.timedelta(days=2),
            ],
            {"min_exclusive": [False, False, True]},
        ),
        (
            cs.Duration(max=dt.timedelta(days=1, seconds=14400)),
            [
                dt.timedelta(seconds=13000),
                dt.timedelta(days=1, seconds=14400),
                dt.timedelta(days=2),
            ],
            {"max": [True, True, False]},
        ),
        (
            cs.Duration(max_exclusive=dt.timedelta(days=1, seconds=14400)),
            [
                dt.timedelta(seconds=13000),
                dt.timedelta(days=1, seconds=14400),
                dt.timedelta(days=2),
            ],
            {"max_exclusive": [True, False, False]},
        ),
    ],
)
def test_validate_min_max(
    column: Column, values: list[Any], valid: dict[str, list[bool]]
):
    tbl = pdt.Table({"a": values})
    actual = evaluate_rules(tbl, column.validation_rules(tbl.a))
    assert actual == valid


@pytest.mark.skipif(C is None, reason="pydiverse.transform is required for this test")
@pytest.mark.parametrize(
    ("column", "values", "valid"),
    [
        (
            cs.Date(resolution="1mo"),
            [dt.date(2020, 1, 1), dt.date(2021, 1, 15), dt.date(2022, 12, 1)],
            {},
            # pydiverse.transform does not support resolution validation
            # resolution should be enforced by convention in pydiverse.pipedag
            # {"resolution": [True, False, True]},
        ),
        (
            cs.Time(resolution="1h"),
            [dt.time(12, 0), dt.time(13, 15), dt.time(14, 0, 5)],
            {},
            # pydiverse.transform does not support resolution validation
            # resolution should be enforced by convention in pydiverse.pipedag
            # {"resolution": [True, False, False]},
        ),
        (
            cs.Datetime(resolution="1d"),
            [
                dt.datetime(2020, 4, 5),
                dt.datetime(2021, 1, 1, 12),
                dt.datetime(2022, 7, 10, 0, 0, 1),
            ],
            {},
            # pydiverse.transform does not support resolution validation
            # resolution should be enforced by convention in pydiverse.pipedag
            # {"resolution": [True, False, False]},
        ),
        (
            cs.Duration(resolution="12h"),
            [
                dt.timedelta(hours=12),
                dt.timedelta(days=2),
                dt.timedelta(hours=5),
                dt.timedelta(hours=12, minutes=30),
            ],
            {},
            # pydiverse.transform does not support resolution validation
            # resolution should be enforced by convention in pydiverse.pipedag
            # {"resolution": [True, True, False, False]},
        ),
    ],
)
def test_validate_resolution(
    column: Column, values: list[Any], valid: dict[str, list[bool]]
):
    tbl = pdt.Table({"a": values})
    actual = evaluate_rules(tbl, column.validation_rules(tbl.a))
    assert actual == valid


@pytest.mark.parametrize(
    "column",
    [
        cs.Datetime(
            min=dt.datetime(2020, 1, 1), max=dt.datetime(2021, 1, 1), resolution="1h"
        )
    ],
)
@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
def test_sample_resolution(column: cs.Column):
    generator = Generator(seed=42)
    samples = column.sample_polars(generator, n=10_000)
    schema = create_colspec("test", {"a": column})
    schema.validate_polars(samples.to_frame("a"))



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause
import pytest

import pydiverse.colspec as cs
from pydiverse.colspec.optional_dependency import C, dy, pdt, pl, validation_mask


class CheckColSpec(cs.ColSpec):
    a = cs.Int64(check=lambda col: (col < 5) | (col > 10))
    b = cs.String(min_length=3, check=lambda col: col.str.contains("x"))


@pytest.mark.skipif(dy.Column is None, reason="dataframely is required for this test")
@pytest.mark.skipif(C is None, reason="pydiverse.transform not installed")
def test_check():
    df = pl.DataFrame({"a": [7, 3, 15], "b": ["abc", "xyz", "x"]})
    _, failures = CheckColSpec.filter_polars(df)
    assert validation_mask(df, failures).to_list() == [False, True, False]
    assert failures.counts() == {"a|check": 1, "b|min_length": 1, "b|check": 1}
    tbl = pdt.Table(df)
    _, failures = CheckColSpec.filter(tbl)
    # assert validation_mask(df, failures).to_list() == [False, True, False]
    assert failures.counts() == {"a|check": 1, "b|min_length": 1, "b|check": 1}



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

import pydiverse.colspec as cs
from pydiverse.colspec.optional_dependency import pyodbc, sa
from pydiverse.colspec.testing import COLUMN_TYPES, create_colspec

try:
    from sqlalchemy.dialects.mssql.pyodbc import MSDialect_pyodbc
    from sqlalchemy.dialects.postgresql.psycopg2 import PGDialect_psycopg2
except ImportError:
    MSDialect_pyodbc = lambda: None  # noqa: E731
    PGDialect_psycopg2 = lambda: None  # noqa: E731


@pytest.mark.skipif(
    sa.Column is None or pyodbc is None,
    reason="sqlalchemy and pyodbc are needed for this test",
)
@pytest.mark.parametrize(
    ("column", "datatype"),
    [
        # (cs.Any(), "SQL_VARIANT"),
        (cs.Bool(), "BIT"),
        (cs.Date(), "DATE"),
        (cs.Datetime(), "DATETIME"),  # consider DATETIME2(6)
        (cs.Time(), "TIME"),  # consider TIME(6)
        (cs.Duration(), "DATETIME"),  # consider DATETIME2(6)
        (cs.Decimal(), "NUMERIC"),
        (cs.Decimal(12), "NUMERIC(12)"),
        (cs.Decimal(None, 8), "NUMERIC(38, 8)"),
        (cs.Decimal(6, 2), "NUMERIC(6, 2)"),
        (cs.Float(), "FLOAT(53)"),
        (cs.Float32(), "FLOAT(24)"),
        (cs.Float64(), "FLOAT(53)"),
        (cs.Integer(), "BIGINT"),
        (cs.Int8(), "SMALLINT"),
        (cs.Int16(), "SMALLINT"),
        (cs.Int32(), "INTEGER"),
        (cs.Int64(), "BIGINT"),
        (cs.UInt8(), "SMALLINT"),  # consider TINYINT
        (cs.UInt16(), "INTEGER"),
        (cs.UInt32(), "BIGINT"),
        (cs.UInt64(), "BIGINT"),
        (cs.String(), "VARCHAR(max)"),
        (cs.String(min_length=3), "VARCHAR(max)"),
        (cs.String(max_length=5), "VARCHAR(5)"),
        (cs.String(min_length=3, max_length=5), "VARCHAR(5)"),
        (cs.String(min_length=5, max_length=5), "CHAR(5)"),
        (cs.String(regex="[abc]de"), "VARCHAR(max)"),
        (cs.String(regex="^[abc]d$"), "VARCHAR(max)"),
        (cs.String(regex="^[abc]{1,3}d$"), "VARCHAR(max)"),
        (cs.Enum(["foo", "bar"]), "CHAR(3)"),
        (cs.Enum(["a", "abc"]), "VARCHAR(3)"),
    ],
)
def test_mssql_datatype(column: cs.Column, datatype: str):
    from sqlalchemy.dialects.mssql.base import MS_2017_VERSION
    from sqlalchemy.dialects.mssql.pyodbc import MSDialect_pyodbc

    dialect = MSDialect_pyodbc()
    dialect.server_version_info = MS_2017_VERSION
    schema = create_colspec("test", {"a": column})
    columns = schema.sql_schema(dialect)
    assert len(columns) == 1
    assert columns[0].type.compile(dialect) == datatype


@pytest.mark.skipif(
    sa.Column is None or pyodbc is None,
    reason="sqlalchemy and pyodbc are needed for this test",
)
@pytest.mark.parametrize(
    ("column", "datatype"),
    [
        (cs.Bool(), "BOOLEAN"),
        (cs.Date(), "DATE"),
        (cs.Datetime(), "TIMESTAMP WITHOUT TIME ZONE"),
        (cs.Time(), "TIME WITHOUT TIME ZONE"),
        (cs.Duration(), "INTERVAL"),
        (cs.Decimal(), "NUMERIC"),
        (cs.Decimal(12), "NUMERIC(12)"),
        (cs.Decimal(None, 8), "NUMERIC(38, 8)"),
        (cs.Decimal(6, 2), "NUMERIC(6, 2)"),
        (cs.Float(), "FLOAT(53)"),
        (cs.Float32(), "FLOAT(24)"),
        (cs.Float64(), "FLOAT(53)"),
        (cs.Integer(), "BIGINT"),
        (cs.Int8(), "SMALLINT"),
        (cs.Int16(), "SMALLINT"),
        (cs.Int32(), "INTEGER"),
        (cs.Int64(), "BIGINT"),
        (cs.UInt8(), "SMALLINT"),
        (cs.UInt16(), "INTEGER"),
        (cs.UInt32(), "BIGINT"),
        (cs.UInt64(), "BIGINT"),
        (cs.String(), "VARCHAR"),
        (cs.String(min_length=3), "VARCHAR"),
        (cs.String(max_length=5), "VARCHAR(5)"),
        (cs.String(min_length=3, max_length=5), "VARCHAR(5)"),
        (cs.String(min_length=5, max_length=5), "CHAR(5)"),
        (cs.String(regex="[abc]de"), "VARCHAR"),
        (cs.String(regex="^[abc]d$"), "VARCHAR"),
        (cs.String(regex="^[abc]{1,3}d$"), "VARCHAR"),
        (cs.Enum(["foo", "bar"]), "CHAR(3)"),
        (cs.Enum(["a", "abc"]), "VARCHAR(3)"),
    ],
)
def test_postgres_datatype(column: cs.Column, datatype: str):
    from sqlalchemy.dialects.postgresql.psycopg2 import PGDialect_psycopg2

    dialect = PGDialect_psycopg2()
    schema = create_colspec("test", {"a": column})
    columns = schema.sql_schema(dialect)
    assert len(columns) == 1
    assert columns[0].type.compile(dialect) == datatype


@pytest.mark.skipif(
    sa.Column is None or pyodbc is None,
    reason="sqlalchemy and pyodbc are needed for this test",
)
@pytest.mark.parametrize("column_type", COLUMN_TYPES)
@pytest.mark.parametrize("nullable", [True, False])
@pytest.mark.parametrize("dialect", [MSDialect_pyodbc()])
def test_sql_nullability(
    column_type: type[cs.Column], nullable: bool, dialect: sa.Dialect
):
    schema = create_colspec("test", {"a": column_type(nullable=nullable)})
    columns = schema.sql_schema(dialect)
    assert len(columns) == 1
    assert columns[0].nullable == nullable


@pytest.mark.skipif(
    sa.Column is None or pyodbc is None,
    reason="sqlalchemy and pyodbc are needed for this test",
)
@pytest.mark.parametrize("column_type", COLUMN_TYPES)
@pytest.mark.parametrize("primary_key", [True, False])
@pytest.mark.parametrize("dialect", [MSDialect_pyodbc(), PGDialect_psycopg2()])
def test_sql_primary_key(
    column_type: type[cs.Column], primary_key: bool, dialect: sa.Dialect
):
    schema = create_colspec("test", {"a": column_type(primary_key=primary_key)})
    columns = schema.sql_schema(dialect)
    assert len(columns) == 1
    assert columns[0].primary_key == primary_key
    assert not columns[0].autoincrement


@pytest.mark.skipif(
    sa.Column is None or pyodbc is None,
    reason="sqlalchemy and pyodbc are needed for this test",
)
@pytest.mark.parametrize("dialect", [MSDialect_pyodbc(), PGDialect_psycopg2()])
def test_sql_multiple_columns(dialect: sa.Dialect):
    schema = create_colspec("test", {"a": cs.Int32(nullable=False), "b": cs.Integer()})
    assert len(schema.sql_schema(dialect)) == 2


@pytest.mark.skipif(
    sa.Column is None or pyodbc is None,
    reason="sqlalchemy and pyodbc are needed for this test",
)
@pytest.mark.parametrize("dialect", [MSDialect_pyodbc(), PGDialect_psycopg2()])
def test_raise_for_list_column(dialect: sa.Dialect):
    # TODO: this probably should raise for MSSQL
    # with pytest.raises(
    #     NotImplementedError, match="SQL column cannot have 'List' type."
    # ):
    cs.List(cs.String()).dtype().to_sql()  # dialect


# @pytest.mark.skipif(sa.Column is None or pyodbc is None, reason="sqlalchemy and
#   pyodbc are needed for this test")
# @pytest.mark.parametrize("dialect", [MSDialect_pyodbc(), PGDialect_psycopg2()])
def test_raise_for_struct_column():  # dialect: sa.Dialect):
    with pytest.raises(
        NotImplementedError, match="Struct column type is not yet implemented"
    ):
        cs.Struct({"a": cs.String()}).dtype()



# Copyright (c) QuantCo and pydiverse contributors 2024-2025
# SPDX-License-Identifier: BSD-3-Clause

import pytest

import pydiverse.colspec as cs
from pydiverse.colspec.columns._base import Column
from pydiverse.colspec.testing.const import ALL_COLUMN_TYPES


@pytest.mark.parametrize("column_type", ALL_COLUMN_TYPES)
def test_string_representation(column_type: type[Column]):
    column = column_type()
    assert str(column) == column_type.__name__.lower()


def test_string_representation_enum():
    column = cs.Enum(["a", "b"])
    assert str(column) == cs.Enum.__name__.lower()


def test_string_representation_list():
    column = cs.List(cs.String())
    assert str(column) == cs.List.__name__.lower()


def test_string_representation_struct():
    column = cs.Struct({"a": cs.String()})
    assert str(column) == cs.Struct.__name__.lower()
